/*********************************************************************
 * remove.c - The "Remove" tab in the package window.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <sys/stat.h>

#define REPORT_REGULAR  0
#define REPORT_SYMLINK  1
#define REPORT_DIR      2
#define REPORT_LOG      3
#define REPORT_ERROR    4
#define REPORT_SHARED   5


/* Check whether a file of a package is shared with any other package. */
static gboolean isshared(gchar *package, gchar *filename)
{
	pkg_t *pkg;
	file_t *file;
	gint n;
	register GSList *p, *f;
	
	for (p = _pkgs; p; p = p->next) {
		pkg = (pkg_t *)(p->data);
		if (!g_ascii_strcasecmp(package, pkg->name))
			continue;
		for (f = pkg->files; f; f = f->next) {
			file = (file_t *)(f->data);
			if (!(n = g_ascii_strcasecmp(filename, file->name)))
				return TRUE;
			else if (n < 0)
				break;
		}
	}
	return FALSE;
}


static void switch_keep_logs(__UNUSED GtkWidget *button, pkg_t *pkg)
{
	g_return_if_fail(pkg != NULL);
	pkg->rm_flags ^= RM_KEEP_LOGS;
}


static void switch_shared(__UNUSED GtkWidget *button, pkg_t *pkg)
{
	g_return_if_fail(pkg != NULL);
	pkg->rm_flags ^= RM_SHARED;
}


static void report(pkg_t *pkg, gchar *file, gint type)
{
	GtkTextBuffer *buf;
	GtkTextIter end;
	gchar *file_8, *msg = NULL, tag[64] = "success";
	
	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->window != NULL);
	
    buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(pkg->rm_textview));
	gtk_text_buffer_get_end_iter(buf, &end);

	file_8 = gp_filename_to_utf8(file);
	
	if (type == REPORT_REGULAR)
		msg = g_strdup_printf(_("Removed '%s'\n"), file_8);
	else if (type == REPORT_SYMLINK)
		msg = g_strdup_printf(_("Removed symlink '%s'\n"), file_8);
	else if (type == REPORT_DIR)
		msg = g_strdup_printf(_("Removed directory '%s'\n"), file_8);
	else if (type == REPORT_LOG)
		msg = g_strdup_printf(_("Removed log '%s'\n"), file_8);
	else if (type == REPORT_ERROR) {
		msg = g_strdup_printf(_("ERROR: '%s': %s.\n"), file_8, g_strerror(errno));
		strcpy(tag, "error");
	}
	else if (type == REPORT_SHARED) {
		msg = g_strdup_printf(_("%s: shared\n"), file_8);
		strcpy(tag, "shared");
	}
	else
		g_assert_not_reached();
	
	g_return_if_fail(pkg->window != NULL);
	
	gtk_text_buffer_insert_with_tags_by_name(buf, &end, msg, -1, tag, "bold", "console", NULL);
	g_free(msg);
	g_free(file_8);
	
	REFRESH_GUI;
	g_return_if_fail(pkg->window != NULL);
	gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(pkg->rm_textview), &end, 0, FALSE, 0, 1);
}


static void removedir(pkg_t *pkg, gchar *path)
{
	gchar *dirname;
	
	g_return_if_fail(path != NULL && *path != '\0');
	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->window != NULL);
	
	if (!strchr(path + 1, '/'))
		return;  /* reached the first level in the filesystem hierarchy */

	if (!g_rmdir(path)) {
		/* the directory was empty and removed successfully */
		report(pkg, path, REPORT_DIR);
		dirname = g_path_get_dirname(path);
		removedir(pkg, dirname);
		g_free(dirname);
	}
	else if (errno != ENOTDIR && errno != ENOENT && errno != ENOTEMPTY)
		report(pkg, path, REPORT_ERROR);
}


static void removelog(pkg_t *pkg, gchar **file)
{
	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->window != NULL);
	
	if (*file == NULL)
		return;
	else if (!g_unlink(*file)) {
		report(pkg, *file, REPORT_LOG);
		*file = NULL;
	}
	else
		report(pkg, *file, REPORT_ERROR);
}


static void remove_pkg(GtkWidget *button, pkg_t *pkg)
{
	file_t *file;
	gint response, count = 0;
	gchar *msg, *dirname, *size_str, *pkgname_8;
	gfloat removed_size = 0.0;
	gboolean remove_log = TRUE;
	GtkWidget *label;
	GtkTextBuffer *buf;
	GtkTextIter start, end;
	register GSList *f;
	struct stat s;

	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->window != NULL);
	
	pkgname_8 = gp_filename_to_utf8(pkg->name);
	msg = g_strdup_printf(_("Remove the package %s ?"), pkgname_8);
	response = gp_question_dialog(pkg->window, msg);
	g_free(msg);
	g_free(pkgname_8);
	if (response != GTK_RESPONSE_YES)
		return;

	gtk_window_set_modal(GTK_WINDOW(pkg->window), TRUE);
	gtk_widget_set_sensitive(button, FALSE);
    buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(pkg->rm_textview));
	gtk_text_buffer_get_end_iter(buf, &end);
	gtk_text_buffer_get_start_iter(buf, &start);
	gtk_text_buffer_delete(buf, &start, &end);
	gtk_text_buffer_insert(buf, &end, "\n", 1);
	
	label = g_object_get_data(G_OBJECT(pkg->window), "remove-label");
	
	/* Remove files */
	for (f = pkg->files; f; f = f->next) {
		
		/* package window destroyed while removing? */
		g_return_if_fail(pkg->window != NULL);
		
		file = (file_t *)(f->data);
		
		/* Skip directories and missing files */
		if ((g_lstat(file->name, &s) < 0) || S_ISDIR(s.st_mode))
			continue;
		
		/* Shared file? */
		else if (!(pkg->rm_flags & RM_SHARED) && isshared(pkg->name, file->name)) {
		 	report(pkg, file->name, REPORT_SHARED);
			remove_log = FALSE;
			continue;
		}
		
		/* Remove the file */
		else if (g_unlink(file->name) < 0) {  /* error */
			report(pkg, file->name, REPORT_ERROR);
			remove_log = FALSE;
			continue;
		}
		
		report(pkg, file->name, S_ISLNK(s.st_mode)? REPORT_SYMLINK : REPORT_REGULAR);
		
		removed_size += file->size;
		size_str = strsize(removed_size);
		
		if (++count == 1)
			msg = g_strdup_printf(_("Removed 1 file  (%s)"), size_str);
		else
			msg = g_strdup_printf(_("Removed %d files  (%s)"), count, size_str);
		
		gtk_label_set_text(GTK_LABEL(label), msg);
		g_free(msg);
		g_free(size_str);

		dirname = g_path_get_dirname(file->name);
		removedir(pkg, dirname);
		g_free(dirname);
	}
	
	update_pkg(pkg);
	
	if (remove_log && !(pkg->rm_flags & RM_KEEP_LOGS)) {
		removelog(pkg, &pkg->log);
		removelog(pkg, &pkg->info);
	}
	else
		gtk_widget_set_sensitive(button, TRUE);
	
	gtk_window_set_modal(GTK_WINDOW(pkg->window), FALSE);
}


GtkWidget *remove_page(pkg_t *pkg)
{
	GtkTextBuffer *buf;  
	GdkColor color;
	GtkWidget *but_remove, *but_shared, *but_keep, *scroll, *vbox, *butbox,
	 *label, *labelbox;
	
	g_return_val_if_fail(pkg != NULL, NULL);
	
	but_shared = gtk_check_button_new_with_mnemonic(_("Remove _shared"));
	gtk_widget_set_sensitive(but_shared, ISWRITABLE(_logdir));
	gtk_tooltips_set_tip(_tips, but_shared, _("Remove also the shared files"
	  " (those which are also owned by other packages)"), NULL);
	g_signal_connect(but_shared, "toggled", G_CALLBACK(switch_shared), pkg);
	 
	but_keep = gtk_check_button_new_with_mnemonic(_("_Keep logs"));
	gtk_widget_set_sensitive(but_keep, ISWRITABLE(_logdir));
	gtk_tooltips_set_tip(_tips, but_keep, _("Don't remove the logs of the package"), NULL);
	g_signal_connect(but_keep, "toggled", G_CALLBACK(switch_keep_logs), pkg);
	 
	but_remove = gp_button(GTK_STOCK_DELETE, _("_Remove"), NULL, NULL);
	gtk_widget_set_sensitive(but_remove, ISWRITABLE(_logdir));
	g_signal_connect(but_remove, "clicked", G_CALLBACK(remove_pkg), pkg);
	
	pkg->rm_textview = gtk_text_view_new();
    gtk_text_view_set_editable(GTK_TEXT_VIEW(pkg->rm_textview), FALSE);
    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(pkg->rm_textview), FALSE);
	gdk_color_parse("black", &color);
	gtk_widget_modify_base(pkg->rm_textview, GTK_STATE_NORMAL, &color);
    g_object_set(pkg->rm_textview, "left-margin", 4, "right-margin", 4, NULL);
    buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(pkg->rm_textview));
	gtk_text_buffer_create_tag(buf, "error", "foreground", "red", NULL);
	gtk_text_buffer_create_tag(buf, "success", "foreground", "white", NULL);
	gtk_text_buffer_create_tag(buf, "shared", "foreground", "lightgreen", NULL);
	gtk_text_buffer_create_tag(buf, "bold", "weight", PANGO_WEIGHT_BOLD, NULL);
	gtk_text_buffer_create_tag(buf, "console", "family", "courier", NULL);
	scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
	 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(scroll), pkg->rm_textview);
	
	label = gtk_label_new(NULL);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	g_object_set_data(G_OBJECT(pkg->window), "remove-label", label);
	
	vbox = gtk_vbox_new(FALSE, 0);
	butbox = gtk_hbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(butbox), but_shared, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(butbox), but_keep, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(butbox), but_remove, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), butbox, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 0);
	labelbox = gtk_hbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(labelbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), labelbox, FALSE, FALSE, 0);
	
	return vbox;
}


/* vim: set ts=4 */
