/*********************************************************************
 * update.c - Update the package database.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <sys/types.h>
#include <sys/stat.h>

 
static pkg_t *get_pkg_from_name(gchar *name)
{
	register GSList *p;
	pkg_t *pkg;
	
	g_return_val_if_fail(name != NULL, NULL);
	
	for (p = _pkgs; p; p = p->next) {
		pkg = (pkg_t *)(p->data);
		if (!strcmp(pkg->name, name))
			return pkg;
	}
	return NULL;
}


static void pkg_row_update(pkg_t *pkg)
{
	guint id;
	gboolean valid;
	GtkTreeIter iter;
	
	g_return_if_fail(pkg != NULL);
	g_return_if_fail((valid = gtk_tree_model_get_iter_first(_model, &iter)));
	
	do {
		gtk_tree_model_get(_model, &iter, COL_ID, &id, -1);
		if (pkg->id == id) {
			main_list_set_pkg(GTK_LIST_STORE(_model), &iter, pkg);
			return;
		}
	}
	while ((valid = gtk_tree_model_iter_next(_model, &iter)));
}


gboolean update_pkg(pkg_t *pkg)
{
	g_return_val_if_fail(pkg != NULL, FALSE);
	
	if (!pkg->log)
		return FALSE;
	if (!ISWRITABLE(_logdir))
		return TRUE;

	g_return_val_if_fail(paco_update(pkg->log, _tmp, _opt & APPARENT_SIZE) == 0, FALSE);
	chmod(pkg->log, 0644);
	
	get_pkg_files(pkg);
	
	if (pkg->window)
		files_treeview_update(pkg);
		
	pkg_row_update(pkg);

	return TRUE;
}


/* Update the packages that are already listed in the main window */
static void update_pkgs(GtkWidget *pbar)
{
	pkg_t *pkg;
	gchar *name, *name_8;
	GtkTreeIter iter;
	gboolean logged, valid;
	G_CONST_RETURN gchar *entry;
	GDir *dir;
	gfloat n = 0.0;
	gint npkgs;
	GSList *li, *unlogged = NULL;
	
	g_return_if_fail((dir = g_dir_open(_logdir, 0, NULL)));
	g_return_if_fail((npkgs = g_slist_length(_pkgs)));
	g_return_if_fail((valid = gtk_tree_model_get_iter_first(_model, &iter)));
	
	while (valid) {
		REFRESH_GUI;
		
		/* Get the name of the package in the row (iter) */
		gtk_tree_model_get(_model, &iter, COL_NAME, &name_8, -1);
		name = gp_filename_from_utf8(name_8);
			
		/* Check wether the package is logged in the log directory */
		for (logged = FALSE; !logged && (entry = g_dir_read_name(dir)); )
			logged = !strcmp(entry, name);
		g_dir_rewind(dir);
		
		pkg = get_pkg_from_name(name);
		
		if (!logged || !update_pkg(pkg)) {
			pkg->log = NULL;
			unlogged = g_slist_prepend(unlogged, (gpointer)pkg);
		}
		
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), n++ / npkgs);
		valid = gtk_tree_model_iter_next(_model, &iter);

		g_free(name);
		g_free(name_8);
	}
	
	g_dir_close(dir);

	/* Remove unlogged packages from the application */
	for (li = unlogged; li; li = li->next)
		destroy_pkg((pkg_t *)(li->data));
	g_slist_free(unlogged);
}


void update_db()
{
	pkg_t *pkg;
	GtkWidget *dialog, *pbar = NULL;
	GDir *dir;
	G_CONST_RETURN gchar *entry;
	gchar *stamp;
	GtkTreeIter iter;
	gint last_id;

	g_return_if_fail(ISWRITABLE(_logdir));
		
	dialog = gp_progress_dialog(_("Updating package database..."), _window.window, &pbar);
	
	/* Update the packages that are already listed in the main window */
	if (_pkgs) {
		update_pkgs(pbar);
		stamp = g_build_filename(_logdir, ".updated", NULL);
		paco_touch(stamp);
		g_free(stamp);
	}
	
	/* Add newly logged packages */
	
	g_return_if_fail((dir = g_dir_open(_logdir, 0, NULL)));
	
	last_id = g_slist_length(_pkgs);
	
	while ((entry = g_dir_read_name(dir))) {
		
		if (ispunct((int)entry[0]))
			continue;
		else if (get_pkg_from_name((gchar *)entry))
			continue;  /* already logged */
			
		_updating = TRUE;
		if ((pkg = new_pkg((gchar *)entry, &last_id))) {
			_pkgs = g_slist_prepend(_pkgs, (gpointer)pkg);
			gtk_list_store_append(GTK_LIST_STORE(_model), &iter);
			main_list_set_pkg(GTK_LIST_STORE(_model), &iter, pkg);
		}	
		_updating = FALSE;
	}
	
	g_dir_close(dir);
	
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), 1.0);
	gtk_widget_destroy(dialog);
}


/* vim: set ts=4 */
