/*********************************************************************
 * window.c - Create the main window, with the list of packages.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <gdk/gdkkeysyms.h>

/* The page displayed in a package window */
#define PAGE_FILES       0
#define PAGE_PROPERTIES  1
#define PAGE_REMOVE      2

static GtkWidget *main_treeview, *about_dialog;


static void about_dialog_destroy()
{
	g_return_if_fail(about_dialog != NULL);
	gtk_widget_destroy(about_dialog);
	about_dialog = NULL;
}


static void about_gpaco()
{
	GtkWidget *label;
	GtkBox *vbox;
	gint i;
	GdkColor color = { 0, 56 * 65535/100, 74 * 65535/100, 56 * 65535/100 };
	gchar *icon_path, *trans = NULL, *markup;
	const gchar *lang, *comment = N_("GTK+ frontend of paco, the\nsource code pacKAGE oRGANIZER");
	const gchar *authors[][2] = {
		{ "ca", "David Rosal" },
		{ "fr", "Ookaze BioACobol" },
		{ "sr", "Boban Petrovic" },
		{ "ru", "Arthur Demchenkov" },
		{ NULL, NULL },
	};
	
	
	if (about_dialog) {
		gtk_window_present(GTK_WINDOW(about_dialog));
		return;
	}

	about_dialog = gtk_dialog_new_with_buttons(
	  _("About gpaco"), GTK_WINDOW(_window.window),
	  GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
	  GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);
	gtk_window_set_resizable(GTK_WINDOW(about_dialog), FALSE);
	gtk_widget_modify_bg(about_dialog, GTK_STATE_NORMAL, &color);
	
	vbox = GTK_BOX(GTK_DIALOG(about_dialog)->vbox);
	 
	icon_path = g_build_filename(DATADIR, "pixmaps", "gpaco.png", NULL);
	gtk_box_pack_start(vbox, gtk_image_new_from_file(icon_path), FALSE, FALSE, 10);
	g_free(icon_path);
	
	label = gtk_label_new(NULL);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_misc_set_padding(GTK_MISC(label), 10, 0);
	
	if ((lang = g_getenv("LC_ALL")) || (lang = g_getenv("LANG"))) {
		for (i = 0; authors[i][0]; i++) {
			if (!strncmp(authors[i][0], lang, 2)) {
				trans = g_strdup_printf(_("\nTranslated by %s"), authors[i][1]);
				break;
			}
		}
	}
	markup = g_markup_printf_escaped(
	 "<span size=\"x-large\" weight=\"heavy\">gpaco %s</span>"
	 "\n\n%s\n\nCopyright &#169; 2004-2005 David Rosal"
	 "%s\n\nhttp://paco.sourceforge.net\n", PACO_VERSION, _(comment), trans ? trans : "");
	gtk_label_set_markup(GTK_LABEL(label), _(markup));
	g_free(markup);
	g_free(trans);
	
	gtk_box_pack_start(vbox, label, FALSE, FALSE, 0);
	g_signal_connect(about_dialog, "response", G_CALLBACK(about_dialog_destroy), NULL);
	gtk_widget_show_all(about_dialog);
}


static gboolean write_main_label(GtkWidget *label)
{
	gfloat total_size = 0.0;
	register GSList *p;
	gchar *msg, *size_str, *logdir_8;
	
	for (p = _pkgs; p; p = p->next) {
		if (p->data)
			total_size += ((pkg_t *)(p->data))->size_inst;
	}
	
	if (total_size == 0) {
		logdir_8 = gp_filename_to_utf8(_logdir);
		msg = g_strdup_printf(_("No packages logged in directory '%s'"), logdir_8);
		g_free(logdir_8);
	}
	else {
		size_str = strsize(total_size);
		msg = g_strdup_printf(_("  %d Packages  |  %s"), g_slist_length(_pkgs), size_str);
		g_free(size_str);
	}
	
	gtk_label_set_text(GTK_LABEL(label), msg);
	g_free(msg);
	
	return TRUE;
}


/* Date printing. */
static void date_cell_data_func(__UNUSED GtkTreeViewColumn *column, GtkCellRenderer *cell,
                                GtkTreeModel *model, GtkTreeIter *iter, gpointer column_)
{
	gint data;
	struct tm *t;
	gchar day[16] = "", hour[8] = "", *txt, *txt_8;
	time_t time__;
	
	gtk_tree_model_get(model, iter, GPOINTER_TO_INT(column_), &data, -1);
	time__ = (time_t)data;
	
	if (!data || !(t = localtime(&time__)))
		txt_8 = g_strdup(" ");
	else {
		if (_opt & SHOW_DAY)
			strftime(day, sizeof(day) - 1, "%d-%b-%Y", t);
		if (_opt & SHOW_HOUR)
			strftime(hour, sizeof(hour) - 1, "%H:%M", t);
		txt = g_strdup_printf("%s %s", day, hour);
		txt_8 = gp_locale_to_utf8(txt);
		g_free(txt);
	}
	
	g_object_set(cell, "text", txt_8, "xpad", 10, NULL);
	g_free(txt_8);
}


static void miss_size_cell_data_func(__UNUSED GtkTreeViewColumn *column, GtkCellRenderer *cell,
                                     GtkTreeModel *model, GtkTreeIter *iter, gpointer col)
{
	gfloat miss_size, size, total_size;
	gchar *txt;
	
	gtk_tree_model_get(model, iter, GPOINTER_TO_INT(col), &miss_size, -1);
	
	if (_opt & MISS_SIZE_AS_PERCENT) {
		gtk_tree_model_get(model, iter, GPOINTER_TO_INT(COL_SIZE), &size, -1);
		if ((total_size = miss_size + size) == 0)
			txt = g_strdup(" ");
		else
			txt = g_strdup_printf("%d%%", (int)(100 * miss_size / total_size));
	}
	else
		txt = strsize(miss_size);
	
	g_object_set(cell, "text", txt, "xalign", 1.0, "xpad", 10, NULL);
	g_free(txt);
}


static void size_cell_data_func(__UNUSED GtkTreeViewColumn *col, GtkCellRenderer *cell,
                                GtkTreeModel *model, GtkTreeIter *iter, gpointer column_)
{
	gfloat data;
	gchar *size_str;
	
	gtk_tree_model_get(model, iter, GPOINTER_TO_INT(column_), &data, -1);
	size_str = strsize(data);
	g_object_set(cell, "text", size_str, "xalign", 1.0, "xpad", 10, NULL);
	g_free(size_str);
}


void main_list_set_pkg(GtkListStore *store, GtkTreeIter *iter, pkg_t *pkg)
{
	GdkPixbuf *pixbuf = NULL;
	gchar *pkgname_8;

	g_return_if_fail(pkg != NULL);
	
	if (pkg->icon)
		pixbuf = gdk_pixbuf_new_from_file_at_size(pkg->icon, 24, 24, NULL);
	
	pkgname_8 = gp_filename_to_utf8(pkg->name);
	
	gtk_list_store_set(store, iter,
	  COL_ICON, pixbuf,
	  COL_NAME, pkgname_8,
	  COL_SIZE, pkg->size_inst,
	  COL_MISSING_SIZE, pkg->size_miss,
	  COL_DATE, pkg->date,
	  COL_FILES, pkg->n_inst,
	  COL_MISSING_FILES, pkg->n_miss,
	  COL_ID, pkg->id, -1);
	
	g_free(pkgname_8);
}


static gint missing_size_sort(GtkTreeModel *model, GtkTreeIter *a,
                              GtkTreeIter *b, __UNUSED gpointer data)
{
	gfloat size_miss_a, size_miss_b, size_a, size_b, total_size_a, total_size_b;
	gint percent_a,	percent_b;
	
	gtk_tree_model_get(model, a, COL_MISSING_SIZE, &size_miss_a, -1);
	gtk_tree_model_get(model, b, COL_MISSING_SIZE, &size_miss_b, -1);

	if (_opt & MISS_SIZE_AS_PERCENT) {
		gtk_tree_model_get(model, a, COL_SIZE, &size_a, -1);
		gtk_tree_model_get(model, b, COL_SIZE, &size_b, -1);
		if (!(total_size_a = size_miss_a + size_a))
			return 1;
		if (!(total_size_b = size_miss_b + size_b))
			return -1;
		percent_a = (gint)(100 * size_miss_a / total_size_a);
		percent_b = (gint)(100 * size_miss_b / total_size_b);
		return percent_a - percent_b;
	}
	else
		return (gint)(size_miss_a - size_miss_b);
}


/* Create a model and fill it with the list of packages. */
static GtkTreeModel *main_list_store_new()
{
	GSList *p;
    GtkListStore *store;
	GtkTreeIter iter;
	
	store = gtk_list_store_new(NUM_COLS,
	  GDK_TYPE_PIXBUF,  /* Icon (pkg->icon) */
	  G_TYPE_STRING,    /* Name of the package (pkg->name) */
	  G_TYPE_FLOAT,     /* Size (pkg->size_inst) */
	  G_TYPE_FLOAT,     /* Missing size (pkg->size_miss) */
	  G_TYPE_INT,       /* Installation date (pkg->date) */
	  G_TYPE_LONG,      /* Number of installed files (pkg->n_inst) */
	  G_TYPE_LONG,      /* Number of missing files (pkg->n_miss) */
	  G_TYPE_INT);      /* id (pkg->id) */

    for (p = _pkgs; p; p = p->next) {
        gtk_list_store_append(store, &iter);
        main_list_set_pkg(store, &iter, (pkg_t *)(p->data));
	}
	
	/* Be case insensitive when sorting by name */
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store), COL_NAME,
	 (GtkTreeIterCompareFunc)casesort, GINT_TO_POINTER(COL_NAME), NULL);
	
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store), COL_MISSING_SIZE,
	 (GtkTreeIterCompareFunc)missing_size_sort, NULL, NULL);
	 
	return GTK_TREE_MODEL(store);
}


static void show_files(pkg_t *pkg)
{
	pkg_window_new(pkg, PAGE_FILES);
}


static void show_properties(pkg_t *pkg)
{
	pkg_window_new(pkg, PAGE_PROPERTIES);
}


static void remove_pkg(pkg_t *pkg)
{
	pkg_window_new(pkg, PAGE_REMOVE);
}


/* XXX Selecci mltiple */
static void remove_log(pkg_t *pkg)
{
	gint res;
	gchar *pkgname_8, *msg;
	
	g_return_if_fail(pkg != NULL);

	pkgname_8 = gp_filename_to_utf8(pkg->name);
	msg = g_strdup_printf(_("Remove the log of the package\n\n%s ?\n"), pkgname_8);
	res = gp_question_dialog(_window.window, msg);
	g_free(msg);
	g_free(pkgname_8);

	if (res == GTK_RESPONSE_YES) {
		gp_unlink(pkg->log);
		pkg->log = NULL;
		gp_unlink(pkg->info);
		pkg->info = NULL;
		destroy_pkg(pkg);
	}
}


/* Switch the rules hint status */
static void switch_rules(GtkWidget *tv)
{
	pkg_t *pkg;
	GSList *li;
	
	_opt ^= SHOW_RULES;
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tv), (_opt & SHOW_RULES));
	
	for (li = _pkgs; li; li = li->next) {
		pkg = (pkg_t *)(li->data);
		if (pkg->window)
			gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(pkg->list_tv), (_opt & SHOW_RULES));
	}
}


/* Switch the visibility of a column on the tree view, and invert the
   corresponding flag in _opt */
static void view_column(gpointer data)
{
	GtkTreeViewColumn *column;
	gboolean visible;
	gint column_id = GPOINTER_TO_INT(data);

	column = gtk_tree_view_get_column(GTK_TREE_VIEW(main_treeview), column_id);
	visible = gtk_tree_view_column_get_visible(column);
	gtk_tree_view_column_set_visible(column, (visible = !visible));
	
	switch (column_id) {
		case COL_ICON:           _opt ^= SHOW_ICON;           return;
		case COL_NAME:           _opt ^= SHOW_NAME;           return;
		case COL_SIZE:           _opt ^= SHOW_SIZE;           return;
		case COL_MISSING_SIZE:   _opt ^= SHOW_MISSING_SIZE;   return;
		case COL_DATE:           _opt ^= SHOW_DATE;           return;
		case COL_FILES:          _opt ^= SHOW_FILES;          return;
		case COL_MISSING_FILES:  _opt ^= SHOW_MISSING_FILES;  return;
	}
}


static pkg_t *get_pkg_from_iter(GtkTreeModel *model, GtkTreeIter *iter)
{
	pkg_t *pkg;
	register GSList *p;
	guint id;
	
	gtk_tree_model_get(model, iter, COL_ID, &id, -1);
	
	for (p = _pkgs; p; p = p->next) {
		pkg = (pkg_t *)(p->data);
		if (pkg->id == id)
			return pkg;
	}
	
	return NULL;
}


static pkg_t *get_selected_pkg(GtkTreeView *tv)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	
	if (gtk_tree_selection_get_selected(gtk_tree_view_get_selection(tv), &model, &iter))
		return get_pkg_from_iter(model, &iter);
		
	return NULL;	
}


static pkg_t *select_pkg(GtkTreeView *tv, GdkEventButton *e)
{
	GtkTreePath *path;
	
	if (gtk_tree_view_get_path_at_pos(tv, e->x, e->y, &path, NULL, 0, 0)) {
		gtk_tree_selection_select_path(gtk_tree_view_get_selection(tv), path);
		return get_selected_pkg(tv);
	}
	return NULL;
}


static void popup_menu(GtkTreeView *tv, GdkEventButton *e)
{
	pkg_t *pkg;
	GtkWidget *menu;

	if (!(pkg = select_pkg(tv, e)))
		return;

	menu = gtk_menu_new();
	
	gp_image_menu_item(GTK_STOCK_JUSTIFY_FILL, _("View _files"), menu, TRUE,
	  G_CALLBACK(show_files), pkg);
	gp_image_menu_item(GTK_STOCK_PROPERTIES, NULL, menu, TRUE, G_CALLBACK(show_properties), pkg);
	gp_image_menu_item(GTK_STOCK_DELETE, _("_Remove..."), menu,
	  ISWRITABLE(_logdir), G_CALLBACK(remove_pkg), pkg);
	SEPARATOR_MENU_ITEM(menu);
	gp_image_menu_item(GTK_STOCK_REFRESH, _("_Update log"), menu,
	  ISWRITABLE(_logdir), G_CALLBACK(update_pkg), pkg);
	gp_image_menu_item(GTK_STOCK_REMOVE, _("_Remove log..."), menu,
	  ISWRITABLE(_logdir), G_CALLBACK(remove_log), pkg);
	SEPARATOR_MENU_ITEM(menu);
	gp_image_menu_item(GTK_STOCK_EXECUTE, _("Create _binary package..."), menu,
	  (_bin_dialog == NULL), G_CALLBACK(binary_pkg), pkg);

	gtk_widget_show_all(menu);
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, e? e->button : 0,
	  gdk_event_get_time((GdkEvent *)e));
}


static gboolean key_press(GtkWidget *tv, GdkEventKey *e, __UNUSED gpointer data)
{
	pkg_t *pkg;
	
	if (!(pkg = get_selected_pkg(GTK_TREE_VIEW(tv))))
		return FALSE;
	else if (e->keyval == GDK_Return)
		show_files(pkg);
	else if (e->keyval == GDK_Delete && ISWRITABLE(_logdir))
		remove_log(pkg);

	return FALSE;
}


static gboolean button_press(GtkWidget *tv, GdkEventButton *e, gpointer data)
{
	GtkTreePath *path;
	GtkTreeIter iter;
	GtkTreeModel *model;

	/* Single click with the right mouse button? */
	if (e->type == GDK_BUTTON_PRESS && e->button == 3)
		popup_menu(GTK_TREE_VIEW(tv), e);
		
	/* Double click with the left mouse button? */
	else if (e->type == GDK_2BUTTON_PRESS && e->button == 1) {
		if (gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(tv), e->x, e->y, &path, NULL, 0, 0)) {
			model = GTK_TREE_MODEL(data);
			if (gtk_tree_model_get_iter(model, &iter, path))
				show_files(get_pkg_from_iter(model, &iter));
		}
	}
	return FALSE;
}


static GtkWidget *main_treeview_new(GtkTreeModel *model)
{
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	GtkWidget *tv;
	
	tv = gtk_tree_view_new_with_model(model);
	g_object_set(tv, "rules-hint", (_opt & SHOW_RULES) != 0, NULL);

	/* Icon column */
	cell = gtk_cell_renderer_pixbuf_new();
	col = gtk_tree_view_column_new_with_attributes(NULL, cell, "pixbuf", COL_ICON, NULL);
	gtk_tree_view_column_set_visible(col, (_opt & SHOW_ICON));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tv), col);
	
	/* Rest of columns */
	gp_column(_("Package"), COL_NAME, (_opt & SHOW_NAME), 0.0, NULL, tv);
	gp_column(_("Size "), COL_SIZE, (_opt & SHOW_SIZE), 1.0,
	  (GtkTreeCellDataFunc)size_cell_data_func, tv);
	gp_column(_("Miss. size "), COL_MISSING_SIZE, (_opt & SHOW_MISSING_SIZE), 1.0,
	  (GtkTreeCellDataFunc)miss_size_cell_data_func, tv);
	gp_column(_("Date "), COL_DATE, (_opt & SHOW_DATE), 0.0,
	  (GtkTreeCellDataFunc)date_cell_data_func, tv);
	gp_column(_("Files"), COL_FILES, (_opt & SHOW_FILES), 0.0, NULL, tv);
	gp_column(_("Miss. files"), COL_MISSING_FILES, (_opt & SHOW_MISSING_FILES), 0.0, NULL, tv);
	
	g_signal_connect(tv, "button-press-event", G_CALLBACK(button_press), (gpointer)model);
	g_signal_connect(tv, "key-press-event", G_CALLBACK(key_press), NULL);
	
	return tv;
}


static void switch_tips()
{
	_opt ^= SHOW_TIPS;
	
	if (_opt & SHOW_TIPS)
		gtk_tooltips_enable(_tips);
	else
		gtk_tooltips_disable(_tips);
}


static gboolean save_window_geometry()
{
	g_return_val_if_fail(_window.window != NULL, FALSE);
	
	gtk_window_get_size(GTK_WINDOW(_window.window), &_window.width, &_window.height);
	gtk_window_get_position(GTK_WINDOW(_window.window), &_window.x, &_window.y);
	
	return FALSE;
}


static GtkWidget *main_menu_bar_new()
{
	GtkWidget *item, *menu, *menubar;
	GtkAccelGroup *accel_group;
	
	g_return_val_if_fail(_window.window != NULL, NULL);
	
	accel_group = gtk_accel_group_new();
    gtk_window_add_accel_group(GTK_WINDOW(_window.window), accel_group);
	g_object_unref(G_OBJECT(accel_group));
	
	menubar = gtk_menu_bar_new();
	
	/* Menu "File" */
	menu = gp_menu(_("_File"), menubar);
	item = gp_image_menu_item(GTK_STOCK_REFRESH, _("_Update database"), menu,
	  ISWRITABLE(_logdir), G_CALLBACK(update_db), NULL);
	gtk_widget_add_accelerator(item, "activate", accel_group, GDK_u,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
	SEPARATOR_MENU_ITEM(menu);
	item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, accel_group);
	g_signal_connect(item, "activate", G_CALLBACK(save_window_geometry), NULL);
	g_signal_connect(item, "activate", gtk_main_quit, NULL);
	gtk_container_add(GTK_CONTAINER(menu), item);
	
	/* Menu "View" */
	menu = gp_menu(_("_View"), menubar);
	gp_check_menu_item(_("_Icon"), (_opt & SHOW_ICON), menu,
      G_CALLBACK(view_column), GINT_TO_POINTER(COL_ICON));
	gp_check_menu_item(_("_Name"), (_opt & SHOW_NAME), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_NAME));
	gp_check_menu_item(_("_Size"), (_opt & SHOW_SIZE), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_SIZE));
	gp_check_menu_item(_("_Missing size"), (_opt & SHOW_MISSING_SIZE), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_MISSING_SIZE));
	gp_check_menu_item(_("_Date"), (_opt & SHOW_DATE), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_DATE));
	gp_check_menu_item(_("_Files"), (_opt & SHOW_FILES), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_FILES));
	gp_check_menu_item(_("Missin_g files"), (_opt & SHOW_MISSING_FILES), menu,
	  G_CALLBACK(view_column), GINT_TO_POINTER(COL_MISSING_FILES));
	SEPARATOR_MENU_ITEM(menu);
	gp_check_menu_item(_("_Rules hint"), (_opt & SHOW_RULES), menu,
	  G_CALLBACK(switch_rules), main_treeview);
	
	/* Menu "Options" */
	menu = gp_menu(_("_Options"), menubar);
	gp_check_menu_item(_("Enable _tips"), (_opt & SHOW_TIPS), menu, G_CALLBACK(switch_tips), NULL);
	SEPARATOR_MENU_ITEM(menu);
	gp_image_menu_item(GTK_STOCK_PREFERENCES, NULL, menu, TRUE, G_CALLBACK(preferences), NULL);

	/* Menu "Search" */
	menu = gp_menu(_("_Search"), menubar);
	item = gp_image_menu_item(GTK_STOCK_FIND, _("Search _file"), menu, TRUE,
	  G_CALLBACK(search), NULL);
	gtk_widget_add_accelerator(item, "activate", accel_group, GDK_f,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
	
	/* Menu "Help" */
	menu = gp_menu(_("_Help"), menubar);
	item = gp_image_menu_item(GTK_STOCK_ABOUT, NULL, menu, TRUE, G_CALLBACK(about_gpaco), NULL);
	gtk_widget_add_accelerator(item, "activate", accel_group, GDK_F1, 0, GTK_ACCEL_VISIBLE);

	return menubar;
}


/* Create the main window. */
void main_window_new()
{
	GtkWidget *scroll, *vbox, *menubar, *label;
	gchar *icon_path;

	_window.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	icon_path = g_build_filename(DATADIR, "pixmaps", "gpaco.png", NULL);
	gtk_window_set_default_icon_from_file(icon_path, NULL);
	g_free(icon_path);
	gtk_window_set_default_size(GTK_WINDOW(_window.window), _window.width, _window.height);
	gtk_window_move(GTK_WINDOW(_window.window), _window.x, _window.y);
	g_signal_connect(_window.window, "delete-event", G_CALLBACK(save_window_geometry), NULL);
	g_signal_connect(_window.window, "destroy", gtk_main_quit, NULL);
	
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(_window.window), vbox);
	
	label = gtk_label_new(NULL);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
	gtk_box_pack_end(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	gtk_box_pack_end(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 0);
	
	_model = main_list_store_new();
	main_treeview = main_treeview_new(_model);
	g_object_unref(_model);
	
	scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(scroll), main_treeview);
	gtk_box_pack_end(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
	
	menubar = main_menu_bar_new();
	gtk_box_pack_end(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);

	g_timeout_add(256, (GSourceFunc)write_main_label, label);
	gtk_widget_show_all(_window.window);
}


/* vim: set ts=4 */
