/**********************************************************************
 * libpaco.c: Handles the system calls that create files and logs them.
 **********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 **********************************************************************/
/* vim: set ts=4 */

#include "config.h"
#include <dlfcn.h>
#include <fcntl.h>			  
#include <stdarg.h>
#include <signal.h>
#include "path.c"

#ifdef ENABLE_NLS
#	include <libintl.h>
#	define _(x) gettext(x)
#else
#	define _(x) x
#endif

#define CHECK_INIT  do { \
	if (!libc_handle || !logfile || !include || !exclude) \
		initialize(); \
} while (0)


static int	(*true_creat)		(const char *, mode_t);
static int	(*true_link)		(const char *, const char *);
static int	(*true_open)		(const char *, int, ...);
static int	(*true_rename)		(const char *, const char *);
static int	(*true_symlink)		(const char *, const char *);
static int	(*true_truncate)	(const char *, off_t);
static FILE	*(*true_fopen)		(const char *, const char *);
static FILE	*(*true_freopen)	(const char *, const char *, FILE *);
#if HAVE_CREAT64
static int	(*true_creat64)		(const char *, mode_t);
#endif
#if HAVE_OPEN64
static int	(*true_open64)		(const char *, int, ...);
#endif
#if HAVE_TRUNCATE64
static int (*true_truncate64)	(const char *, off64_t);
#endif
#if HAVE_FOPEN64
static FILE	*(*true_fopen64)	(const char *, const char *);
#endif
#if HAVE_FREOPEN64
static FILE	*(*true_freopen64)	(const char *, const char *, FILE *);
#endif

static char *logfile, *include, *exclude;
static int enable_debug;
static void *libc_handle;


#define FILTER(path) (!paco_inpaths(path, DEFAULT_EXCLUDE))

		
__attribute__((__format__(printf, 1, 2)))
static void debug(const char *fmt, ...)
{
	va_list a;

	if (enable_debug) {
		fflush(stdout);
		fprintf(stderr, "paco-log >> ");
		va_start(a, fmt);
		vfprintf(stderr, fmt, a);
		va_end(a);
		fputc('\n', stderr);
	}
}


__attribute__((__noreturn__,__format__(printf, 1, 2)))
static void die_errno(const char *fmt, ...)
{
	va_list a;
	
	fflush(stdout);
	fprintf(stderr, "paco-log: ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
	fprintf(stderr, ": %s\n", strerror(errno));
	
	exit(EXIT_FAILURE);
}


__attribute__((__noreturn__,__format__(printf, 1, 2)))
static void die(const char *fmt, ...)
{
	va_list a;
	
	fflush(stdout);
	fprintf(stderr, "paco-log: ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
	fputc('\n', stderr);
	
	exit(EXIT_FAILURE);
}


__attribute__((__noreturn__))
static RETSIGTYPE segfault(__UNUSED int sig)
{
	die(_("Segmentation fault\nPlease report this bug to <%s>"),
	 PACKAGE_BUGREPORT);
}


static void *xdlsym(const char *symbol)
{
	void *ret;
	char *error;
	
	dlerror();  /* clear errors */
	ret = dlsym(libc_handle, symbol);
	if ((error = (char *)dlerror()))
		die("dlsym(%p, \"%s\"): %s", libc_handle, symbol, error);
	
	return ret;
}										   


static char *xgetenv(const char *var)
{
	char *val;
	
	if (!(val = getenv(var)))
		die(_("variable %s undefined"), var);
	
	return val;
}


static void initialize()
{
	char *aux;
	
	if (libc_handle)
		return;

	if ((aux = getenv("PACO_DEBUG")))
		enable_debug = !strcmp(aux, "yes");
		
	debug("initialize()");
	
	signal(SIGSEGV, segfault);

	dlerror();  /* clear errors */

#ifdef RTLD_NEXT
	libc_handle = RTLD_NEXT;
#else
	libc_handle = dlopen(LIBC__SO, RTLD_LAZY);
#endif

	if (libc_handle == NULL)
		die("%s", dlerror());

	logfile = xgetenv("PACO_TMPFILE");
	include = xgetenv("PACO_INCLUDE");
	exclude = xgetenv("PACO_EXCLUDE");

	true_creat 		= xdlsym("creat");
	true_link 		= xdlsym("link");
	true_open 		= xdlsym("open");
	true_rename 	= xdlsym("rename");
	true_symlink 	= xdlsym("symlink");
	true_truncate 	= xdlsym("truncate");
	true_fopen		= xdlsym("fopen");
	true_freopen	= xdlsym("freopen");
#if HAVE_OPEN64
	true_open64 	= xdlsym("open64");
#endif
#if HAVE_CREAT64
	true_creat64 	= xdlsym("creat64");
#endif
#if HAVE_TRUNCATE64
	true_truncate64 = xdlsym("truncate64");
#endif
#if HAVE_FOPEN64
			true_fopen64	= xdlsym("fopen64");
#endif
#if HAVE_FREOPEN64
	true_freopen64	= xdlsym("freopen64");
#endif
}


static void log_file(const char *path)
{
	int fd, len;
	char real[PATH_MAX];
	
	CHECK_INIT;
	
	if ((len = paco_realpath(path, real, 1)) < 0)
		die_errno("paco_realpath(\"%s\", %p, 1)", path, real);
	
	if (!FILTER(real))
		return;
	else if (!paco_inpaths(real, include) || paco_inpaths(real, exclude)) {
		debug(_("%s: excluded."), real);
		return;
	}

	if ((fd = true_open(logfile, O_WRONLY | O_CREAT | O_APPEND, 0644)) < 0)
		die_errno("open(\"%s\", O_WRONLY | O_CREAT | O_APPEND, 0644)", logfile);
	else if (write(fd, real, len) != len)
		die_errno("write(%d, \"%s\", %d)", fd, real, len);
	else if (write(fd, "\n", 1) != 1)
		die_errno("write(%d, \"\n\", 1)", fd);
	else if (close(fd) < 0)
		die_errno("close(%d)", fd);
	
	debug(_("%s: logged."), real);
}


/*=== System call handlers ===*/


FILE *fopen(const char *path, const char *mode)
{
	FILE *ret;
	
	CHECK_INIT;
	
	ret = true_fopen(path, mode);
	
	if (FILTER(path) && memchr("wa+", (int)*mode, 3)) {
		debug("fopen(\"%s\", \"%s\")", path, mode);
		log_file(path);
	}
	
	return ret;
}


FILE *freopen(const char *path, const char *mode, FILE *stream)
{
	FILE *ret;
	
	CHECK_INIT;
	
	ret = true_freopen(path, mode, stream);
	
	if (FILTER(path) && memchr("wa+", (int)*mode, 3)) {
		debug("freopen(\"%s\", \"%s\")", path, mode);
		log_file(path);
	}
	
	return ret;
}


int rename(const char *oldpath, const char *newpath)
{
	int ret;
	
	CHECK_INIT;

	ret = true_rename(oldpath, newpath);
	
	if (FILTER(newpath)) {
		debug("rename(\"%s\", \"%s\")", oldpath, newpath);
		log_file(newpath);
	}
	
	return ret;
}


int creat(const char *path, mode_t mode)
{
	int ret;
	
	CHECK_INIT;

	ret = true_open(path, O_CREAT | O_WRONLY | O_TRUNC, mode);
	
	if (FILTER(path)) {
		debug("creat(\"%s\", 0%o)", path, (int)mode);
		log_file(path);
	}
	
	return ret;
}


int link(const char *oldpath, const char *newpath)
{
	int ret;
	
	CHECK_INIT;

	ret = true_link(oldpath, newpath);

	if (FILTER(newpath)) {
		debug("link(\"%s\", \"%s\")", oldpath, newpath);
		log_file(newpath);
	}
	
	return ret;
}


int truncate(const char *path, off_t length)
{
	int ret;
	
	CHECK_INIT;

	ret = true_truncate(path, length);
	
	if (FILTER(path)) {
		debug("truncate(\"%s\", %d)", path, (int)length);
		log_file(path);
	}
	
	return ret;
}


int open(const char *path, int flags, ...)
{
	va_list a;
	mode_t mode;
	int ret;
	
	CHECK_INIT;
	
	va_start(a, flags);
	mode = va_arg(a, mode_t);
	va_end(a);
	
	ret = true_open(path, flags, mode);
	
	if (FILTER(path) && (flags & (O_WRONLY | O_RDWR | O_CREAT | O_APPEND))) {
		debug("open(\"%s\")", path);
		log_file(path);
	}
	
	return ret;
}


int symlink(const char *oldpath, const char *newpath)
{
	int ret;
	
	CHECK_INIT;

	ret = true_symlink(oldpath, newpath);
	
	if (FILTER(newpath)) {
		debug("symlink(\"%s\", \"%s\")", oldpath, newpath);
		log_file(newpath);
	}
	
	return ret;
}


#if HAVE_CREAT64
int creat64(const char *path, mode_t mode)
{
	int ret;
	
	CHECK_INIT;

	ret = true_open64(path, O_CREAT | O_WRONLY | O_TRUNC, mode);

	if (FILTER(path)) {
		debug("creat64(\"%s\")", path);
		log_file(path);
	}
	
	return ret;
}
#endif


#if HAVE_OPEN64
int open64(const char *path, int flags, ...)
{
	va_list a;
	mode_t mode;
	int ret;
	
	CHECK_INIT;
	
	va_start(a, flags);
	mode = va_arg(a, mode_t);
	va_end(a);
	
	ret = true_open64(path, flags, mode);
	
	if (FILTER(path) && (flags & (O_WRONLY | O_RDWR | O_CREAT | O_APPEND))) {
		debug("open64(\"%s\")", path);
		log_file(path);
	}
	
	return ret;
}
#endif


#if HAVE_TRUNCATE64
int truncate64(const char *path, off64_t length)
{
	int ret;
	
	CHECK_INIT;

	ret = true_truncate64(path, length);

	if (FILTER(path)) {
		debug("truncate64(\"%s\", %d)", path, (int)length);
		log_file(path);
	}
	
	return ret;
}
#endif


#if HAVE_FOPEN64
FILE *fopen64(const char *path, const char *mode)
{
	FILE *ret;
	
	CHECK_INIT;

	ret = true_fopen64(path, mode);
	
	if (FILTER(path) && memchr("wa+", (int)*mode, 3)) {
		debug("fopen64(\"%s\", \"%s\")", path, mode);
		log_file(path);
	}
	
	return ret;
}
#endif


#if HAVE_FREOPEN64
FILE *freopen64(const char *path, const char *mode, FILE *stream)
{
	FILE *ret;
	
	CHECK_INIT;

	ret = true_freopen64(path, mode, stream);
	
	if (FILTER(path) && memchr("wa+", (int)*mode, 3)) {
		debug("freopen64(\"%s\", \"%s\")", path, mode);
		log_file(path);
	}
	
	return ret;
}
#endif
