/*********************************************************************
 * util.c - Utilities for both paco and gpaco
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/
/* vim: set ts=4 */

#include "util.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>

#ifdef ENABLE_NLS
#	include <libintl.h>
#	define _(x) gettext(x)
#else
#	define _(x) x
#endif

/* For *BSD statfs() */
#ifdef HAVE_SYS_PARAM_H
#	include <sys/param.h>
#endif
#if HAVE_SYS_MOUNT_H
#	include <sys/mount.h>
#endif

#if (HAVE_SYS_STATVFS_H && HAVE_STATVFS)
#	define STATFS statvfs
#	include <sys/statvfs.h>
#elif HAVE_STATFS
#	define STATFS statfs
#	if HAVE_SYS_VFS_H
#		include <sys/vfs.h>
#	elif HAVE_SYS_STATFS
#		include <sys/statfs.h>
#	endif
#elif defined STATFS
#	undef STATFS
#endif


void paco_die(const char *fmt, ...)
{
	va_list a;

	fflush(stdout);
	fprintf(stderr, "paco: ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
	fprintf(stderr, "\n");
	
	exit(EXIT_FAILURE);
}


void paco_die_errno(const char *fmt, ...)
{
	va_list a;

	fflush(stdout);
	fprintf(stderr, "paco: ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
	fprintf(stderr, ": %s\n", strerror(errno));
	
	exit(EXIT_FAILURE);
}


/* fgets() wrapper that strip trailing newlines, tabs and spaces. */
char *paco_fgets(char *buf, int size, FILE *f)
{
	register char *p;
	register int c;
	
	p = buf;
	
	while (--size) {
		if ((c = fgetc(f)) == EOF)
			break;
		else if ((*p++ = c) == '\n')
			break;
	}
	
	if (p == buf)
		return NULL;
		
	for (*p = '\0'; isspace((int)*(--p)); ) {
		*p = '\0';
		if (p == buf)
			break;
	}

	return buf;
}


static int strcnt(char *str, const char *word)
{
	char *p;
	int cnt = 0;
	
	for (p = str; *p && (p = strstr(p, word)); cnt++, p++);
	return cnt;
}


/* Expand environment variable HOME in str */
static char *expand_home(char *str)
{
	char *home, *p, *ret;
	int i, home_len, h, hh;

	h = strcnt(str, "$HOME");
	hh = strcnt(str, "${HOME}");
		
	if (!(home = getenv("HOME")) || (h + hh == 0))
		return strdup(str);

	home_len = strlen(home);
	if (!(ret = malloc(strlen(str) + hh * (home_len - 5) + h * (home_len - 7))))
		paco_die("malloc(): %s", _("Out of memory"));
	
	for (i = 0, p = str; *p; ) {
		if (!strncmp(p, "$HOME", 5)) {
			memcpy(ret + i, home, home_len);
			i += home_len;
			p += 5;
		}
		else if (!strncmp(p, "${HOME}", 7)) {
			memcpy(ret + i, home, home_len);
			i += home_len;
			p += 7;
		}
		else
			*(ret + i++) = *p++;
	}
	*(ret + i) = '\0';
	
	return ret;
}


int paco_getconf(const char *file, const char *tag, void *var, int type,
                 int flag)
{
	FILE *f;
	char buf[PATH_MAX];
	size_t n, ret = 0;

	assert(tag != NULL && var != NULL);
	
	if (!(f = fopen(file, "r")))
		return 0;

	n = strlen(tag);
	
	while (paco_fgets(buf, sizeof(buf), f) && !ret) {
		
		if (strncmp(buf, tag, n) || buf[n++] != '=')
			continue;
		
		switch (type) {
			case PACO_INT:
				*(int *)var = atoi(&buf[n]);
				break;
			case PACO_FLAGS:
	            if (buf[n] == '1')
	                *(int *)var |= flag;
	            else if (buf[n] == '0')
	                *(int *)var &= ~flag;
				break;
			case PACO_STRING:
				*(char **)var = expand_home(&buf[n]);
				break;
			default:
				assert(type == PACO_INT || type == PACO_FLAGS
				  || type == PACO_STRING);
		}
		ret++;
	}
	
	fclose(f);
	return ret;
}


long paco_getsize(int apparent, const char *fmt, ...)
{
	struct stat s;
	va_list a;
	char path[PATH_MAX];
	
#ifdef STATFS
	struct STATFS f;
	int blocks;
#endif
 	
	va_start(a, fmt);
	vsnprintf(path, sizeof(path), fmt, a);
	va_end(a);
	
	/* missing file */
	if (lstat(path, &s) < 0)
		return SIZE_MISSING;
	
	/* symbolic link */
	else if (S_ISLNK(s.st_mode))
		return SIZE_SYMLINK;
	
	/* regular file */
#ifdef STATFS
	else if (!apparent && STATFS(path, &f) == 0 && f.f_bsize > 0) {
		/* round up the size to the nearest multiple of the block size */
		blocks = (s.st_size / f.f_bsize) + ((s.st_size % f.f_bsize) != 0);
		return (long)(blocks * f.f_bsize);
	}
#endif
	else  /* less correct value */
		return (long)s.st_size;
}


void paco_touch(const char *path)
{
	FILE *f;

	if ((f = fopen(path, "w"))) {
		fprintf(f, "%s\n", PACO_VERSION);
		fclose(f);
	}
}


FILE *paco_fopen(const char *path, const char *mode)
{
	FILE *ret = fopen(path, mode);
	if (!ret)
		paco_die_errno("fopen(\"%s\", \"%s\")", path, mode);
	return ret;
}


static int __paco_open(const char *path, int flags)
{
	int fd = open(path, flags);
	if (fd < 0)
		paco_die_errno("open(\"%s\")", path);
	return fd;
}


static void brutecopy(const char *source, const char *dest)
{
	int n, fd1, fd2;
	char buf[KILOBYTE << 2];
	
	fd1 = __paco_open(source, O_RDONLY);
	fd2 = __paco_open(dest, O_CREAT | O_WRONLY | O_TRUNC);
		
	while ((n = read(fd1, buf, sizeof(buf)))) {
		if (n < 0)
			paco_die_errno("read()");
		else if (write(fd2, buf, n) < 0)
			paco_die_errno("write(%d, \"%s\", %d)", fd2, buf, n);
	}
	if (close(fd1) < 0)
		paco_die_errno("close(%d)", fd1);
	else if (close(fd2) < 0)
		paco_die_errno("close(%d)", fd2);
}


/* Wrapper around rename(), that deals with files in different filesystems. */
void paco_rename(const char *oldpath, const char *newpath)
{
	if (rename(oldpath, newpath) < 0) {
		if (errno != EXDEV)
			paco_die_errno("rename(\"%s\", \"%s\")", oldpath, newpath);
		brutecopy(oldpath, newpath);
	}
}


int paco_update(const char *log, const char *tmp_log, int apparent)
{
	FILE *flog, *ftmp;
	int found;
	char buf[PATH_MAX], *p, *k, *file;
	long gz, bz2, raw, old_raw, old_gz, old_bz2;

	flog = paco_fopen(log, "r");
	ftmp = paco_fopen(tmp_log, "w");
	
	/* header */
	fgets(buf, sizeof(buf), flog);
	assert(!strncmp(buf, "#!paco", 6));
	fputs(buf, ftmp);
	
	if (!paco_fgets(buf, sizeof(buf), flog)) {
		/* no logged files */
		fclose(flog);
		fclose(ftmp);
		unlink(log);
		unlink(tmp_log);
		return PACO_ERROR_EMPTY_LOG;
	}

	do {
		
		if (!(k = strchr(buf, '/')))
			continue;

		else if ((p = strchr(k, '|')))
			*p++ = '\0';

		file = strdup(k);
		
		found = ((raw = paco_getsize(apparent, "%s", file)) != -2);
		found += ((gz = paco_getsize(apparent, "%s.gz", file)) != -2);
		found += ((bz2 = paco_getsize(apparent, "%s.bz2", file)) != -2);
		
		if (found)
			fprintf(ftmp, "%s|%ld|%ld|%ld\n", file, raw, gz, bz2);
		else if (p && sscanf(p, "%ld|%ld|%ld", &old_raw, &old_gz, &old_bz2) == 3)
			fprintf(ftmp, "-%s|%ld|%ld|%ld\n", file, old_raw, old_gz, old_bz2);
		else
			fprintf(ftmp, "-%s|-2|-2|-2\n", file);

		free(file);
	}
	while (paco_fgets(buf, sizeof(buf), flog));
	
	fclose(flog);
	fclose(ftmp);
	
	paco_rename(tmp_log, log);
	
	return 0;
}
