/*********************************************************************
 * info.c - Display package information.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/
/* vim: set ts=4: */

#include "paco.h"
#include <glob.h>
#include <sys/stat.h>


/* Pair of variable/value given by a %define in the .spec or .pc file */
typedef struct {
	char *var;
	char *val;
	char fmt;
#define FMT_SPEC '%'
#define FMT_PC   '$'
} define_t;

#define SPEC   0
#define ASPEC  1

typedef struct {
	char *cc;
	char *cflags;
	char *ldflags;
	char *cppflags;
	char *cxx;
	char *cxxflags;
	char *name;
	char *version;
	char *author;
	char *summary;
	char *url;
	char *license;
	char *desc;
	char *conf_opts;
	char *icon;
} info_t;

static info_t *info;
static char **pkg_dirs;
static int info_cnt;


/* Parse the .spec file and store defines */
static define_t *get_spec_defines(char *file)
{
	define_t *def = NULL;
	char *var, *p, *val, buf[KILOBYTE];
	int k = 0;
	FILE *f;
	
	if (!(f = fopen(file, "r")))
		return NULL;
		
	while (fgets(buf, sizeof(buf), f)) {
		if (!(p = strtok(buf, " \t")) || strcmp(p, "%define"))
			continue;
		if (!(var = strtok(NULL, " \t\n")))
			continue;
		if (!(val = strtok(NULL, " \t\n")))
			continue;
		def = xrealloc(def, (k + 2) * sizeof(define_t));
		def[k].fmt = FMT_SPEC;
		def[k].var = strdup(var);
		def[k++].val = strdup(val);
	}
	if (k)
		def[k].var = NULL;
		
	fclose(f);
	return def;
}


/* Get defines from a pkg-config .pc file */
static define_t *get_pc_defines(char *file)
{
	define_t *def = NULL;
	char *c, buf[KILOBYTE];
	int k = 0;
	FILE *f;
	
	if (!(f = fopen(file, "r")))
		return NULL;
		
	while (paco_fgets(buf, sizeof(buf), f)) {
		if ((c = strchr(buf, '=')) && *(c + 1)) {
			*c++ = '\0';
			def = xrealloc(def, (k + 2) * sizeof(define_t));
			def[k].fmt = FMT_PC;
			def[k].var = strdup(buf);
			def[k++].val = strdup(c);
		}
	}
	if (k)
		def[k].var = NULL;
		
	fclose(f);
	return def;
}


/* Resolve defines in 'buf' */
static char *resolve_defines(char *buf, define_t *def)
{
	int i, k, n, len;
	size_t j = 0;
	static char res[KILOBYTE];

	if (!def || !strchr(buf, def[0].fmt))
		return buf;
		
	for (i = j = len = 0; (buf[i] && j < sizeof(res)); ) {
		if (buf[i] == def[0].fmt) {
			if (buf[++i] == '{')
				i++;
			for (k = 0; def[k].var; k++) {
				len = strlen(def[k].var);
				if (!strncmp(buf + i, def[k].var, len))
					break;
			}
			if (def[k].var) {
				for (n = 0; (def[k].val[n] && j < sizeof(res)); n++)
					res[j++] = def[k].val[n];
				i += len + (buf[i + len] == '}');
				continue;
			}
			else if (buf[--i] == '{')
				i--;
		}
		res[j++] = buf[i++];
	}
	res[j] = '\0';
	
	return res;
}


static char *get_info(char *file, const char *tag, define_t *def)
{
	FILE *f;
	char buf[KILOBYTE], *p, *val;
	int len;
	
	if (!(f = fopen(file, "r")))
		return NULL;
		
	len = strlen(tag);
	
	while (paco_fgets(buf, sizeof(buf), f)) {
		p = def ? resolve_defines(buf, def) : buf;
		if (!strncasecmp(p, tag, len) && *(p += len)) {
			val = p + strspn(p, " \t:=");
			/* avoid things like 'Version: @VERSION@' */
			if (*val == '@' && *(val + strlen(val) - 1) == '@')
				continue;
			fclose(f);
			info_cnt++;
			return strdup(val);
		}
	}
	
	fclose(f);
	return NULL;
}


static char *get_description(char *spec_file, const char *title, define_t *def)
{
	size_t nl = 1, len, i = 0, ok = 0, title_len;
	char desc[KILOBYTE << 3], buf[KILOBYTE], *p;
	FILE *f;
	
	memset(&desc, 0, sizeof(desc));

	title_len = strlen(title);
	
	if (!(f = fopen(spec_file, "r")))
		return NULL;
	
	while (!ok && paco_fgets(buf, sizeof(buf), f))
		ok = (!strncmp(buf, title, title_len));
	if (!ok) {
		fclose(f);
		return NULL;
	}
	
	while (paco_fgets(buf, sizeof(buf), f)
	  && buf[0] != title[0] && buf[0] != '#') {
	
		if (buf[0] == '\0') {
			if (!nl) {
				if (i + 1 >= sizeof(desc))
					break;
				desc[i++] = '\n';
				nl++;
			}
		}
		else {
			p = def ? resolve_defines(buf, def) : buf;
			len = strlen(p) + 1;
			if (i + len >= sizeof(desc))
				break;
			strcat(desc, p);
			if (len < 48) {
				strcat(desc, "\n");
				nl++;
			}
			else {
				strcat(desc, " ");
				nl = 0;
			}
			i += len;
		}
	}
	if (!nl && i < sizeof(desc))
		strcat(desc, "\n");

	fclose(f);
	
	if (i > 8) {
		info_cnt++;
		return strdup(desc);
	}
	return NULL;
}

	
static void free_defines(define_t **def)
{
	int i;
	
	if (*def == NULL)
		return;
	for (i = 0; (*def + i)->var; i++) {
		free((*def + i)->var);
		xfree((*def + i)->val);
	}
	free(*def);
}


static void get_info_aspec(char *aspec_file)
{
	info->name = get_info(aspec_file, "ShortName", NULL);
	info->version = get_info(aspec_file, "PackageVersion", NULL);
	info->url = get_info(aspec_file, "URL", NULL);
	info->license = get_info(aspec_file, "License", NULL);
	info->desc = get_description(aspec_file, "[Description]", NULL);
	if (!(info->author = get_info(aspec_file, "Maintainer", NULL)))
		info->author = get_info(aspec_file, "Packager", NULL);
	if (!(info->summary = get_info(aspec_file, "Summary", NULL)))
		info->summary = get_info(aspec_file, "DisplayName", NULL);
}


static void get_info_spec(char *spec_file)
{
	define_t *def;
	
	def = get_spec_defines(spec_file);
	
	info->name = get_info(spec_file, "Name", def);
	info->version = get_info(spec_file, "Version", def);
	info->summary = get_info(spec_file, "Summary", def);
	info->url = get_info(spec_file, "URL", def);
	if (!(info->author = get_info(spec_file, "Vendor", def)))
		info->author = get_info(spec_file, "Packager", def);
	if (!(info->license = get_info(spec_file, "License", def)))
		info->license = get_info(spec_file, "Copyright", def);
	info->desc = get_description(spec_file, "%description", def);
	
	free_defines(&def);
}


static void get_info_desktop(char *desktop_file)
{
	info->name = get_info(desktop_file, "Name", NULL);
	if (!(info->summary = get_info(desktop_file, "Comment", NULL)))
		info->summary = get_info(desktop_file, "GenericName", NULL);
}


static void get_info_pc(char *pc_file)
{
	define_t *def;
	
	def = get_pc_defines(pc_file);
	
	info->name = get_info(pc_file, "Name", def);
	info->version = get_info(pc_file, "Version", def);
	info->summary = get_info(pc_file, "Description", def);

	free_defines(&def);
}


static char *getvar(char *file, const char *tag)
{
	FILE *f;
	char buf[PATH_MAX];
	size_t n, m;

	if (!file || !tag || !(f = fopen(file, "r")))
		return NULL;
		
	n = strlen(tag);
	
	while (paco_fgets(buf, sizeof(buf), f)) {
		if (!strncmp(buf, tag, n) && (m = strspn(&buf[n], " =\t"))) {
			fclose(f);
			return strdup(&buf[n + m]);
		}
	}

	fclose(f);
	return NULL;
}


/* Extract the value of 'var' from Makefile, makefile or config.log */
static char *get_build_var(const char *var)
{
	char *val;
	
	if (!(val = getvar("Makefile", var)) || !*val) {
		if (!(val = getvar("makefile", var)) || !*val) {
			if (!(val = getvar("config.log", var)) || !*val)
				return NULL;
		}
	}
	return val;
}


static char *search_file(const char *name, const char *suffix)
{
	char *str, *base, *ret = NULL, *pattern[4];
	glob_t pglob;
	int i, j, k, retval;
	
	memset(&pglob, 0, sizeof(glob_t));

	/* strip the version suffix from the name of the package */
	base = strdup(name);
	for (i = 0; base[i]; i++) {
		if (ispunct((int)base[i]) && isdigit((int)base[i + 1]))
			break;
	}
	base[i] = '\0';
	
	debug(1, _("searching %s%s... "), base, suffix ? suffix : "");
	
	xasprintf(&pattern[0], "*/*/*/%s%s", base, suffix ? suffix : "");
	xasprintf(&pattern[1], "*/*/*/%s-[0-9]*%s", base, suffix ? suffix : "");
	xasprintf(&pattern[2], "%s.in", pattern[0]);
	xasprintf(&pattern[3], "%s.in", pattern[1]);

	for (i = 0; i < 4; i++) {
		for (j = 0; pkg_dirs[j]; j++) {
			for (k = 3; k >= 0; k--) {
				xasprintf(&str, "%s/%s", pkg_dirs[j], &pattern[i][k * 2]);
				retval = glob(str, 0, NULL, &pglob);
				free(str);
				if (retval == 0 && pglob.gl_pathv[0] && pglob.gl_pathv[0][0]) {
					ret = strdup(pglob.gl_pathv[0]);
					goto Return;
				}
			}
		}
	}
	
	Return:
	for (i = 0; i < 4; i++)
		free(pattern[i]);
	globfree(&pglob);
	xfree(base);
	
	if (ret)
		debug(0, _("found %s.\n"), ret);
	else
		debug(0, _("not found.\n"));
		
	return ret;
}


static char *get_env(const char *var)
{
	char *aux;
	
	if ((aux = getenv(var))) {
		info_cnt++;
		return strdup(aux);
	}
	return NULL;
}


static char *get_config_status_opts(char *config_status)
{
	FILE *f;
	char buf[KILOBYTE], *p;

	if (!(f = fopen(config_status, "r")))
		return NULL;

	while (paco_fgets(buf, sizeof(buf), f) && buf[0] == '#') {
		if (strstr(buf, "/configure")) {
			fclose(f);
			p = strstr(buf, "--");
			info_cnt++;
			return strdup(p ? p : "");
		}
	}
	
	fclose(f);
	return NULL;
}


static char *get_config_log_opts(char *config_log)
{
	FILE *f;
	char buf[KILOBYTE], *p;

	if (!(f = fopen(config_log, "r")))
		return NULL;

	while (paco_fgets(buf, sizeof(buf), f)) {
		if ((p = strchr(buf, '$')) && strstr(p, "/configure")) {
			fclose(f);
			p = strstr(buf, "--");
			info_cnt++;
			return strdup(p ? p : "");
		}
	}
	
	fclose(f);
	return NULL;
}


static char *get_icon(char *pkgname, char **desktop_file)
{
	char buf[KILOBYTE], *s, *ikon, *icon = NULL, *pkg_log;
	char *fmt[] = IMAGE_FORMATS;
	FILE *flog;
	int ok;
	size_t i;
	
	*desktop_file = search_file(pkgname, ".desktop");
	
	if (*desktop_file && (icon = getvar(*desktop_file, "Icon"))) {
		for (i = ok = 0; !ok && fmt[i]; i++)
			ok += (strstr(icon, fmt[i]) != NULL);
		if (!ok) {
			free(icon);
			icon = NULL;
		}
	}
	if (!icon) {
		for (i = 0; fmt[i] && !icon; i++)
			icon = search_file(pkgname, fmt[i]);
		if (!icon)
			return NULL;
	}
	
	xasprintf(&pkg_log, "%s/%s", _logdir, pkgname);
	flog = fopen(pkg_log, "r");
	free(pkg_log);
	if (!flog)
		return NULL;
	
	if ((ikon = strrchr(icon, '/')))
		ikon++;
	else
		ikon = icon;
		
	while (paco_fgets(buf, sizeof(buf), flog)) {
		if (buf[0] != '#' && (s = strstr(buf, ikon)) && *(s - 1) == '/') {
			s += strlen(ikon);
			if (*s == '|' || *s == '\0') {
				*s = '\0';
				fclose(flog);
				free(icon);
				info_cnt++;
				return strdup(strchr(buf, '/'));
			}
		}
	}
	fclose(flog);
	free(icon);

	return NULL;
}


static void print_debug_info()
{
	int width, i;
	
	width = get_screen_width() - strlen("paco >> ");
	
	debug(1, "-");
	for (i = 1; i < width; i++)
		fputc('-', stderr);
	fputc('\n', stderr);
	
	if (info->name)
		debug(1, "Name:     %s\n", info->name);
	if (info->version)
		debug(1, "Version:  %s\n", info->version);
	if (info->author)
		debug(1, "Author:   %s\n", info->author);
	if (info->summary)
		debug(1, "Summary:  %s\n", info->summary);
	if (info->url)
		debug(1, "URL:      %s\n", info->url);
	if (info->license)
		debug(1, "License:  %s\n", info->license);
	if (info->desc)
		debug(1, "Descr:    (%d chars written)\n", (int)strlen(info->desc));
	if (info->conf_opts)
		debug(1, "Opts:     %s\n", info->conf_opts);
	if (info->cc)
		debug(1, "CC:       %s\n", info->cc);
	if (info->cflags)
		debug(1, "CFLAGS:   %s\n", info->cflags);
	if (info->ldflags)
		debug(1, "LDFLAGS   %s\n", info->ldflags);
	if (info->cppflags)
		debug(1, "CPPFLAGS: %s\n", info->cppflags);
	if (info->cxx)
		debug(1, "CXX:      %s\n", info->cxx);
	if (info->cxxflags)
		debug(1, "CXXFLAGS: %s\n", info->cxxflags);
	if (info->icon)
		debug(1, "Icon:     %s\n", info->icon);
	
	debug(1, "-");
	for (i = 1; i < width; i++)
		fputc('-', stderr);
	fputc('\n', stderr);
}


/* Get package information and write it into the 'info log' of the package. */
void getinfo(char *pkg_name)
{
	char *file, *desktop_file, *info_dir = NULL, *info_file = NULL;
	FILE *f;
	
	pkg_dirs = xmalloc0(4 * sizeof(char *));
	pkg_dirs[0] = strdup(".");
	pkg_dirs[1] = get_build_var("top_srcdir");
	pkg_dirs[pkg_dirs[1] ? 2 : 1] = get_build_var("top_builddir");
	pkg_dirs = unics(pkg_dirs);

	info = xmalloc0(sizeof(info_t));
	
	debug(1, _("---- Searching information:\n"));

	info->cc = get_env("CC");
	info->cflags = get_env("CFLAGS");
	info->ldflags = get_env("LDFLAGS");
	info->cppflags = get_env("CPPFLAGS");
	info->cxx = get_env("CXX");
	info->cxxflags = get_env("CXXFLAGS");
	
	info->icon = get_icon(pkg_name, &desktop_file);
	
	if ((file = search_file("config.log", NULL)))
		info->conf_opts = get_config_log_opts(file);
	
	xfree(file);
	
	if (!info->conf_opts && (file = search_file("config.status", NULL))) {
		info->conf_opts = get_config_status_opts(file);
		xfree(file);
	}
	
	if ((file = search_file(pkg_name, ".spec")))
		get_info_spec(file);
	else if ((file = search_file(pkg_name, ".aspec")))
		get_info_aspec(file);
	else if ((file = search_file(pkg_name, ".pc")))
		get_info_pc(file);
	else if (desktop_file)
		get_info_desktop(desktop_file);
	
	if (!info_cnt)
		goto Return;
	
	xasprintf(&info_dir, "%s/_info", _logdir);
	if (mkdir(info_dir, 0755) < 0 && errno != EEXIST)
		die_errno("mkdir(\"%s\", 0755)", info_dir);
	xasprintf(&info_file, "%s/%s", info_dir, pkg_name);
	
	f = paco_fopen(info_file, "w+");
	
	fprintf(f, "#!paco-%s\n", PACO_VERSION);

	if (info->conf_opts)
		fprintf(f, "#c:%s\n", info->conf_opts);
	if (info->cc)
		fprintf(f, "#e:CC=%s\n", info->cc);
	if (info->cflags)
		fprintf(f, "#e:CFLAGS=%s\n", info->cflags);
	if (info->ldflags)
		fprintf(f, "#e:LDFLAGS=%s\n", info->ldflags);
	if (info->cppflags)
		fprintf(f, "#e:CPPFLAGS=%s\n", info->cppflags);
	if (info->cxx)
		fprintf(f, "#e:CXX=%s\n", info->cxx);
	if (info->cxxflags)
		fprintf(f, "#e:CXXFLAGS=%s\n", info->cxxflags);
	if (info->icon)
		fprintf(f, "#i:%s\n", info->icon);
	if (info->name)
		fprintf(f, "Name:     %s\n", info->name);
	if (info->version)
		fprintf(f, "Version:  %s\n", info->version);
	if (info->author)
		fprintf(f, "Author:   %s\n", info->author);
	if (info->summary)
		fprintf(f, "Summary:  %s\n", info->summary);
	if (info->url)
		fprintf(f, "URL:      %s\n", info->url);
	if (info->license)
		fprintf(f, "License:  %s\n", info->license);
	if (info->desc)
		fprintf(f, "\nDescription\n%s\n", info->desc);

	fclose(f);

	if (_verbosity >= PACO_DEBUG)
		print_debug_info();
		
	xfree(info->cc);
	xfree(info->cflags);
	xfree(info->ldflags);
	xfree(info->cppflags);
	xfree(info->cxx);
	xfree(info->cxxflags);
	xfree(info->name);
	xfree(info->version);
	xfree(info->author);
	xfree(info->summary);
	xfree(info->url);
	xfree(info->license);
	xfree(info->desc);
	xfree(info->conf_opts);
	xfree(info->icon);
	xfree(info);

	Return:
	xfree(info_dir);
	xfree(file);
	xfree(info_file);
	xfree(desktop_file);
	xfree2(pkg_dirs);
}


static void query_file(void *data, void *user_data)
{
	list_t *f;
	file_t *file;
	pkg_t *pkg = (pkg_t *)data;
	char *path = (char *)user_data;
	
	for (f = pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (!strcmp(file->name, path)) {
			_exit_status = EXIT_SUCCESS;
			printf("\t%s", pkg->name);
			break;
		}
	}
}


/* Get the pkg(s) that own a given file, and print the results. */
static void query(char *path, list_t *pkgs)
{
	char real[PATH_MAX];
	 
	if (paco_realpath(path, real, 0) < 0)
		die_errno("paco_realpath(\"%s\", %p, 0)", path, real);
		
	printf("%s:", real);
	list_foreach(pkgs, query_file, (void *)real);
	putchar('\n');
}


void do_query(char *args[])
{
	int i;
	list_t *pkgs;
	
	pkgs = get_pkgs(ALL_PACKAGES);	
	get_files(pkgs, FILE_LOGGED);
	
	_exit_status = EXIT_FAILURE;
	
	for (i = 0; args[i]; i++)
		query(args[i], pkgs);
	
	pkg_list_free(pkgs);
}


/* This lovely function takes a string buffer and prints it out in pieces not
 * longer than a given maximmum length. The cut points are not in the middle
 * of words if possible.
 */
static void cutprint(char *buf, size_t cut)
{
    size_t i = 0, j = 0, len = 0;
    char out[KILOBYTE];
    int beggining = 1, /* Leading spaces indicator */
        spaces = 0;    /* True if there are any space to cut at */
 
    /* If the buffer is short enough, just print it out and return */
    if (strlen(buf) <= cut) {
        printf("%s", buf);
        return;
    }
    /* Split the buffer into pieces of maximmum length = 'cut' */
    while (buf[i]) {
        if (buf[i] == ' ')
            spaces = !beggining;
        else 
            beggining = 0;
        if (len < cut) {
            len++;
            out[j++] = buf[i++];
            continue;
        }
        /* len >= cut */
        if (spaces && !beggining) {
            for (; buf[i] != ' '; j--, i--);
            i++;
        }
        /* Split the buffer and print the piece */
		out[j] = '\0';
        j = len = spaces = 0;
        puts(out);
        while (buf[i] == ' ')
			i++;
    }
	
    /* The remaining piece */
    out[j] = '\0';
    printf("%s", out);
}


static int print_info(pkg_t *pkg, int screen_width)
{
	FILE *f;
	char buf[PATH_MAX];
	int k, cnt = 0;
	
	if (!pkg->info || !(f = fopen(pkg->info, "r")))
		return 0;
		
	while (paco_fgets(buf, sizeof(buf), f)) {
		if (buf[0] == '#')
			continue;
		else if (cnt++ == 0) {
			k = strlen(pkg->name) + 2;
			putnchar('-', k);
			printf("\n %s \n", pkg->name);
			putnchar('-', k);
			putchar('\n');
		}
		cutprint(buf, screen_width);
		putchar('\n');
	}

	fclose(f);
	return cnt;
}


static void print_vars(char *info_file, char *vars[])
{
	char *tok, *val = NULL;
	int i;

	for (i = 0; vars[i]; i++) {
		xasprintf(&tok, "#e:%s", vars[i]);
		if ((val = get_info(info_file, tok, NULL))) {
			printf("%s=%s\n", vars[i], val);
			free(val);
		}
		free(tok);
	}
}


void do_var(char *args[], char *env_vars[])
{
	list_t *p, *pkgs;
	pkg_t *pkg;
	int print_name, i, all = 0;
	char *all_vars[] = ALLVARS;
	
	pkgs = get_pkgs(args);
	print_name = (list_length(pkgs) > 1 || _verbosity);

	for (i = 0; !all && env_vars[i]; i++)
		all = !strcmp(env_vars[i], "all");
	
	for (p = pkgs; p; p = p->next) {
		pkg = (pkg_t *)p->data;
		if (print_name)
			printf("%s:\n", pkg->name);
		print_vars(pkg->info, all ? all_vars : env_vars);
		putnchar('\n', (p->next != NULL));
	}
	
	pkg_list_free(pkgs);
	xfree2(env_vars);
}


void do_conf(char *args[])
{
	list_t *pkgs, *p;
	pkg_t *pkg;
	int print_name;
	char *opts;

	pkgs = get_pkgs(args);
	print_name = (list_length(pkgs) > 1 || _verbosity);

	for (p = pkgs; p; p = p->next) {
		pkg = (pkg_t *)p->data;
		if (print_name)
			printf("%s:\n", pkg->name);
		if ((opts = get_info(pkg->info, "#c", NULL)) && *opts)
			puts(opts);
		putnchar('\n', (p->next != NULL));
		xfree(opts);
	}
	
	pkg_list_free(pkgs);
}


void do_info(char *args[])
{
	list_t *pkgs, *p;
	int screen_width;
	
	pkgs = get_pkgs(args);
	screen_width = get_screen_width();

	for (p = pkgs; p; p = p->next) {
		if (print_info((pkg_t *)p->data, screen_width) && p->next)
			putchar('\n');
	}
	
	pkg_list_free(pkgs);
}
