/********************************************************************
 * list.c - List packages/files with different levels of detail.
 ********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 ********************************************************************/

#include "paco.h"

/* Used in this file */
static unsigned long total_shared;
static list_opts_t *opt;

struct {
	int size;
	int size_inst;
	int size_miss;
	int n_inst;
	int n_miss;
	int n_shared;
} width;


static int sort_pkgs_by_size_inst(list_t *p, list_t *q)
{
	return (((pkg_t *)p->data)->size_inst > ((pkg_t *)q->data)->size_inst);
}


static int sort_pkgs_by_date(list_t *p, list_t *q)
{
	return (((pkg_t *)p->data)->date < ((pkg_t *)q->data)->date);
}


static int sort_pkgs_by_n_inst(list_t *p, list_t *q)
{
	return (((pkg_t *)p->data)->n_inst > ((pkg_t *)q->data)->n_inst);
}


static int sort_pkgs_by_n_miss(list_t *p, list_t *q)
{
	return (((pkg_t *)p->data)->n_miss > ((pkg_t *)q->data)->n_miss);
}


static int sort_pkgs_by_size_miss(list_t *p, list_t *q)
{
	return (((pkg_t *)p->data)->size_miss > ((pkg_t *)q->data)->size_miss);
}


static int sort_pkgs_by_name(list_t *p, list_t *q)
{
	return strcasecmp(((pkg_t *)q->data)->name, ((pkg_t *)p->data)->name);
}


/* size printing */
static char *strsize(long size, int width)
{
	static char str[64];
	
	width += !width;
	
	if (size == SIZE_SYMLINK)
		snprintf(str, sizeof(str), "%*c", width, '@');
	else if (opt->blksize == HUMAN_READABLE) { 
		if (size < KILOBYTE)
			snprintf(str, sizeof(str), "%*ld", width, size);
		else if (size < MEGABYTE)
			snprintf(str, sizeof(str), "%*ldk", --width, size / KILOBYTE);
		else if (size < (10 * MEGABYTE))
			snprintf(str, sizeof(str), "%*.1fM", --width, (float)size/MEGABYTE);
		else
			snprintf(str, sizeof(str), "%*ldM", --width, size / MEGABYTE);
	}
	else
		snprintf(str, sizeof(str), "%*ld", width, (long)(size / opt->blksize));
		
	return str;
}


/* Nice date printing. */
static char *strdate(int time_)
{
	static char ret[32] = { 0 };
	struct tm *t;
	time_t time__ = (time_t)time_;
	
	if (!time_ || !(t = localtime(&time__)))
		sprintf(ret, "%*c", (opt->p_flags & PKG_HOUR) ? 17 : 11, ' ');
	
	else if (opt->p_flags & PKG_HOUR)
		strftime(ret, sizeof(ret) - 1, "%d-%b-%Y %H:%M", t);
	else
		strftime(ret, sizeof(ret) - 1, "%d-%b-%Y", t);
		
	return ret;
}


/* Return the number of digits of a number. */
__attribute__((__const__))
static int digits(long n)
{
	int ret;
	for (ret = 0; n; n /= 10, ret++);
	return ret;
}


/* Calculate the minimmum width required to display the size. */
static int size_digits(unsigned long size)
{
	if (opt->blksize == HUMAN_READABLE)
		return ((size / MEGABYTE) > 9999) ? 7 : 5;
	else	
		return digits(size / opt->blksize);
}


/* Whether 'my_file' is logged by 'pkg' */
static int isloggedby(file_t *my_file, pkg_t *pkg, int set_shared)
{
	list_t *f;
	file_t *file;
	int n;
	
	for (f = pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (!(n = strcmp(file->name, my_file->name))) {
			if (set_shared) {
				my_file->shared++;
				file->shared++;
			}
			return 1;
		}
		else if (n > 0)
			break;
	}
	return 0;
}


static void print_file(void *data, __UNUSED void *user_data)
{
	file_t *file = (file_t *)data;
	char ln[PATH_MAX];
	int cnt;
		
	if (opt->f_flags & FILE_SIZE)
		printf("%s  %s", strsize(file->size, width.size), file->name);
	else
		printf("%s", file->name);

	if (file->size == SIZE_SYMLINK && (opt->f_flags & FILE_SYMLINKS)) {
		if ((cnt = readlink(file->name, ln, sizeof(ln) - 1)) > 0)
			ln[cnt] = '\0';
		else
			memcpy(ln, "?", 2);
		printf(" -> %s", ln);
	}
	
	putchar('\n');
}


static void list_pkg_files(pkg_t *pkg)
{
	if ((opt->f_flags & FILE_NO_PKGNAME) == 0)
		printf("%s:\n", pkg->name);
	
	if (opt->sort == SORT_SIZE)
		pkg->files = list_sort(pkg->files, sort_files_by_size);
	else
		pkg->files = list_sort(pkg->files, sort_files_by_name);
		
	if (opt->flags & LIST_REVERSE)
		pkg->files = list_reverse(pkg->files);
		
	list_foreach(pkg->files, print_file, NULL);
}


static void print_shared_files(void *data, void *user_data)
{
	list_t *f;
	file_t *file;
	pkg_t *my_pkg, *pkg;
	int k = 0;
	
	my_pkg = (pkg_t *)user_data;
	pkg = (pkg_t *)data;
	
	if (!strcmp(pkg->name, my_pkg->name))
		return;
		
	for (f = my_pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (!isloggedby(file, pkg, 0))
			continue;
		else if (opt->f_flags & FILE_SHARER) {
			if (!k++)
				printf("(%s)\n", pkg->name);
			print_file(file, NULL);
		}
		else if (!file->shared++)
			/* avoid printing the file more than once */
			print_file(file, NULL);
	}
}	


static void print_shared(void *data, void *user_data)
{
	pkg_t *pkg;
	list_t *all_pkgs;
	
	pkg = (pkg_t *)data;
	all_pkgs = (list_t *)user_data;
	
	if ((opt->f_flags & FILE_NO_PKGNAME) == 0)
		printf("%s:\n", pkg->name);
	
	list_foreach(all_pkgs, print_shared_files, (void *)pkg);
}
	

static void list_shared_files(list_t *pkgs)
{
	list_t *all_pkgs;
	
	all_pkgs = get_pkgs(ALL_PACKAGES);
	get_files(all_pkgs, opt->f_type);
	
	list_foreach(all_pkgs, sort_pkg_files_by_name, NULL);
	list_foreach(pkgs, print_shared, (void *)all_pkgs);
	
	pkg_list_free(all_pkgs);
}


void do_list_files(char *args[], list_opts_t *op_list)
{
	list_t *p, *pkgs;
	total_t total;
	
	opt = op_list;
	
	pkgs = get_pkgs(args);
	get_files(pkgs, opt->f_type);
	
	if (opt->f_flags & FILE_SIZE) {
		memset(&total, 0, sizeof(total));
		get_totals(pkgs, &total);
		width.size = size_digits(total.size_inst + total.size_miss);
	}

	if (opt->f_type & FILE_SHARED)
		list_shared_files(pkgs);
	else {
		for (p = pkgs; p; p = p->next) {
			list_pkg_files((pkg_t *)(p->data));
			if ((opt->f_flags & FILE_NO_PKGNAME) == 0 && p->next)
				putchar('\n');
		}	
		if ((opt->f_flags & FILE_SIZE) && opt->flags & LIST_TOTAL) {
			printf("\n%s: %s\n", _("TotaL"),
			 strsize(total.size_inst + total.size_miss, 0));
		}
	}

	pkg_list_free(pkgs);
}


static unsigned long get_n_shared(pkg_t *my_pkg, list_t *all_pkgs)
{
	list_t *p, *f;
	pkg_t *pkg;
	file_t *file;
	unsigned long n_shared = 0;
	
	for (f = my_pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (file->shared) {
			n_shared++;
			continue;
		}
		for (p = all_pkgs; p; p = p->next) {
			pkg = (pkg_t *)(p->data);
			if (!strcmp(my_pkg->name, pkg->name))
				continue;
			if (isloggedby(file, pkg, 1)) {
				n_shared++;
				break;
			}
		}
	}
	total_shared += n_shared;
	return n_shared;
}


static void list_pkg(void *data, void *user_data)
{
	pkg_t *pkg;
	list_t *all_pkgs;
	unsigned long n_shared;

	pkg = (pkg_t *)data;
	all_pkgs = (list_t *)user_data;
	
	if (opt->p_flags & PKG_SIZE_INST) {
		printf("%s ", strsize(pkg->size_inst, width.size_inst));
		putnchar(' ', !(opt->p_flags & PKG_SIZE_MISS));
	}
	if (opt->p_flags & PKG_SIZE_MISS) { 
		if (pkg->size_miss)
			printf("[%s]  ", strsize(pkg->size_miss, width.size_miss));
		else
			printf("[%*c]  ", width.size_miss, ' ');
	}
	if (opt->p_flags & PKG_N_INST) {
		printf("%*ld ", width.n_inst, pkg->n_inst);
		putnchar(' ', !(opt->p_flags & (PKG_N_MISS | PKG_N_SHARED)));
	}
	if (opt->p_flags & PKG_N_MISS) {
		if (pkg->n_miss)
			printf("[%*ld] ", width.n_miss, pkg->n_miss);
		else
			printf("[%*c] ", width.n_miss, ' ');
		putnchar(' ', !(opt->p_flags & PKG_N_SHARED));
	}
	if (opt->p_flags & PKG_N_SHARED) {
		if ((n_shared = get_n_shared(pkg, all_pkgs)))
			printf("(%*ld)  ", width.n_shared, n_shared);
		else
			printf("(%*c)  ", width.n_shared, ' ');
	}
	if (opt->p_flags & PKG_DATE) 	
		printf("%s  ", strdate(pkg->date));
	puts(pkg->name);
}


static void print_pkg_totals(total_t *total)
{
	if (opt->p_flags & PKG_SIZE_INST) {
		printf("%s ", strsize(total->size_inst, width.size_inst));
		putnchar(' ', !(opt->p_flags & PKG_SIZE_MISS));
	}
	if (opt->p_flags & PKG_SIZE_MISS) {
		if (total->size_miss)
			printf("[%s]  ", strsize(total->size_miss, width.size_miss));
		else
			printf("[%*c]  ", width.size_miss, ' ');
	}
	if (opt->p_flags & PKG_N_INST) {
		printf("%*ld ", width.n_inst, total->n_inst);
		putnchar(' ', !(opt->p_flags & (PKG_N_MISS | PKG_N_SHARED)));
	}
	if (opt->p_flags & PKG_N_MISS) {
		if (total->n_miss)
			printf("[%*ld] ", width.n_miss, total->n_miss);
		else
			printf("[%*c] ", width.n_miss, ' ');
		putnchar(' ', !(opt->p_flags & PKG_N_SHARED));
	}
	if (opt->p_flags & PKG_N_SHARED) {
		if (total_shared)
			printf("(%*ld)  ", width.n_shared, total_shared);
		else
			printf("(%*c)  ", width.n_shared, ' ');
	}
	if (opt->p_flags & PKG_DATE)
		printf("%*c", (opt->p_flags & PKG_HOUR) ? 19 : 13, ' ');
	puts(_("TotaL"));
}


/* Print the names of the pkgs in a 'ls' style (in columns) */
static void ls_pkgs(list_t *pkgs)
{
	int count = 0, i = 0, j, rows, cols, pacs;
	size_t maxlen = 0;
	char **name;
	list_t *p;
	pkg_t *pkg;
	
	pacs = list_length(pkgs);
	name = xmalloc(pacs * sizeof(char *));
	
	for (p = pkgs; p; p = p->next) {
		pkg = (pkg_t *)(p->data);
		maxlen = MAX(maxlen, strlen(pkg->name));
		name[i++] = pkg->name;
	}
	
	maxlen += 2;
	if ((cols = get_screen_width() / maxlen) == 0)
		cols = 1;
	rows = (pacs / cols) + ((pacs % cols) != 0);

	if (rows == 1) {
		for (j = 0; j < pacs; j++) {
			if (count++)
				printf("  ");
			printf("%s", name[j]);
		}
		putchar('\n');
	}
	else {
		for (i = 0; i < rows; i++) {
			for (j = i; j < pacs; j += rows)
				printf("%s%*c", name[j], (int)(maxlen - strlen(name[j])), ' ');
			putchar('\n');
		}
	}
	free(name);
}	

void do_list_pkgs(char *args[], list_opts_t *op_list)
{
	list_t *pkgs, *all_pkgs = NULL;
	total_t total;
	list_cmp_f *cmp;

	pkgs = get_pkgs(args);
	opt = op_list;
	
	if (opt->p_flags & (PKG_SIZE | PKG_N)
	  || (opt->sort && opt->sort != SORT_DATE)) {
		memset(&total, 0, sizeof(total));
		get_files(pkgs, FILE_LOGGED);
		get_totals(pkgs, &total);
	}

	switch (opt->sort) {
		case SORT_SIZE:		cmp = sort_pkgs_by_size_inst; break;
		case SORT_DATE:		cmp = sort_pkgs_by_date; break;
		case SORT_N_FILES:	cmp = sort_pkgs_by_n_inst; break;
		case SORT_N_MFILES:	cmp = sort_pkgs_by_n_miss; break;
		case SORT_MSIZE:	cmp = sort_pkgs_by_size_miss; break;
		default:			cmp = sort_pkgs_by_name;
	}
	pkgs = list_sort(pkgs, cmp);
	if (opt->flags & LIST_REVERSE)
		pkgs = list_reverse(pkgs);

	if (opt->p_flags & PKG_SIZE_INST)
		width.size_inst = size_digits(total.size_inst);
	if (opt->p_flags & PKG_SIZE_MISS)
		width.size_miss = size_digits(total.size_miss);
	if (opt->p_flags & PKG_N_MISS)
		width.n_miss = digits(total.n_miss);
	if (opt->p_flags & PKG_N_INST)
		width.n_inst = digits(total.n_inst);
	if (opt->p_flags & PKG_N_SHARED) {
		width.n_shared = digits(MAX(total.n_inst, total.n_miss));
		all_pkgs = get_pkgs(ALL_PACKAGES);
		get_files(all_pkgs, FILE_LOGGED);
		list_foreach(all_pkgs, sort_pkg_files_by_name, NULL);
	}
	if (opt->p_flags > 0) {
	 	list_foreach(pkgs, list_pkg, (void *)all_pkgs);
		if ((opt->flags & LIST_TOTAL) && list_length(pkgs) > 1
		 && (opt->p_flags & (PKG_SIZE | PKG_N))) {
		 	print_pkg_totals(&total);
		}
	}
	else
		ls_pkgs(pkgs);
	
	pkg_list_free(all_pkgs);
	pkg_list_free(pkgs);
}


/* vim: set ts=4: */
