/******************************************************************
 * log.c - Log the installation of a package
 ******************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 ******************************************************************/

#include "paco.h"
#include <glob.h>
#include <sys/stat.h>

#ifdef HAVE_SYS_WAIT_H
#	include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
#	define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
#	define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#define EXIT_FAILURE_EXTERNAL	2


/* Search for libpaco-log.so in the filesystem.
 * Take into account libpaco-log.so.0.1.0, libpaco-log.so.0.0 and the like. */
static char *search_libpaco_log()
{
	char *libpath, *pattern;
	glob_t pglob;
	
	xasprintf(&libpath, "%s/libpaco-log.so", LIBDIR);
	
	if (!access(libpath, X_OK))
		return libpath;
	
	memset(&pglob, 0, sizeof(glob_t));
	
	xasprintf(&pattern, "%s.[0-9]*", libpath);
	
	if (!glob(pattern, 0, NULL, &pglob) && pglob.gl_pathc) {
		free(libpath);
		libpath = strdup(pglob.gl_pathv[0]);
	}
	
	free(pattern);
	globfree(&pglob);
	
	return libpath;
}


/* Remove duplicated non-commented lines from an unsorted file. */
static int unique(char *file)
{
	FILE *f1, *f2, *ft;
	char tmpfile[KILOBYTE], buf1[KILOBYTE], buf2[KILOBYTE];
	int n = 0, rep;
	
	ft = tmp_fopen(tmpfile, sizeof(tmpfile));
	f1 = paco_fopen(file, "r");
	f2 = paco_fopen(file, "r");
	
	while (fgets(buf1, sizeof(buf1), f1)) {
		if (buf1[0] == '#') {
			fputs(buf1, ft);
			continue;
		}
		fseek(f2, ftell(f1), SEEK_SET);
		for (rep = 0; !rep && fgets(buf2, sizeof(buf2), f2); )
			rep = !strcmp(buf1, buf2);
		if (!rep) {
			fputs(buf1, ft);
			n++;
		}
	}
	
	fclose(f1);
	fclose(f2);
	fclose(ft);

	paco_rename(tmpfile, file);
	
	return n ? 0 : PACO_ERROR_EMPTY_LOG;
}


static void log_stdin(log_opts_t *opt)
{
	char buf[PATH_MAX];
	FILE *ftmp;
	
	ftmp = paco_fopen(_tmp, "w");
	
	while (paco_fgets(buf, sizeof(buf), stdin)) {
		if (buf[0] == '#' || buf[0] == '\0' || buf[0] == '\n')
			continue;
		else if (!paco_inpaths(buf, DEFAULT_EXCLUDE)
		  && !paco_inpaths(buf, opt->exclude) && paco_inpaths(buf, opt->include)) {
			fprintf(ftmp, "%s\n", buf);
		}
	}
	
	fclose(ftmp);
}


/* Set environment variable */
#if HAVE_SETENV
static void setvar(char *var, char *val)
{
	if (setenv(var, val, 1) < 0)
		die_errno("setenv(\"%s\", \"%s\", 1)", var, val);
	debug(1, "%s=%s\n", var, val); 
}
#else
static void setvar(char *var, char *val)
{
	char *str;
	
	xasprintf(&str, "%s=%s", var, val);
	if (putenv(str) < 0)
		die_errno("putenv(\"%s\")", str);
	debug(1, "%s=%s\n", var, val); 
}
#endif


static void log_cmd(char *args[], log_opts_t *opt)
{
	int i, status;
	char *cmd[4];
	
	switch (fork()) {
		case 0:	
			/* Build the command to execute */
			cmd[0] = strdup("sh");
			cmd[1] = strdup("-c");
			cmd[2] = strdup(args[0]);
			for (i = 1; args[i]; i++)
				xasprintf(&cmd[2], "%s %s", cmd[2], args[i]);
			cmd[3] = NULL;
			
			/* Set environment variables */
			setvar("PACO_TMPFILE", _tmp);
			setvar("PACO_INCLUDE", opt->include);
			setvar("PACO_EXCLUDE", opt->exclude);
			setvar("LD_PRELOAD", search_libpaco_log());
			if (_verbosity >= PACO_DEBUG)
				setvar("PACO_DEBUG", "yes");

			/* Execute the command */
			execv("/bin/sh", cmd);
			die_errno("execv()");
		
		case -1:
			die_errno("fork()");
	}
	
	wait(&status);

	if (opt->ignore_errors)
		return;
	else if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS)
		exit(EXIT_FAILURE_EXTERNAL);
}


static void createdir(char *dir)
{
	struct stat st;
	char *cmd;

	if (lstat(dir, &st) < 0) {
		if (mkdir(dir, 0755) < 0) {
			xasprintf(&cmd, "mkdir --mode=0755 --parents %s", dir);
			system(cmd);
			free(cmd);
		}
	}
	else if (!S_ISDIR(st.st_mode))
		die(_("%s: Not a directory"), dir);
}
	
	
/* Remove .gz or .bz2 suffix from a string */
static char *zstrip(char *buf)
{
	size_t n = strlen(buf);
	
	if (n > 4 && !memcmp(buf + n - 4, ".bz2", 4))
		buf[n - 4] = '\0';
	else if (n > 3 && !memcmp(buf + n - 3, ".gz", 3))
		buf[n - 3] = '\0';
	
	return buf;
}


static int filter_file(const char *file)
{
	struct stat s;
	
	if (lstat(file, &s) < 0 || S_ISDIR(s.st_mode))
		return 0;
		
	debug(1, _("logged %s\n"), file);
	return 1;
}


static void log_stdout()
{
	char buf[PATH_MAX];
	FILE *ftmp;
	
	/* Open '_tmp' and select the valid entries */
	ftmp = paco_fopen(_tmp, "r");
	
	while (paco_fgets(buf, sizeof(buf), ftmp)) {
		if (filter_file(buf))
			printf("%s\n", buf);
	}
	
	fclose(ftmp);
}


static void log_pkg(log_opts_t *opt)
{
	char buf[PATH_MAX], *log, *info_log;
	FILE *f, *ftmp;

	createdir(_logdir);
	xasprintf(&log, "%s/%s", _logdir, opt->pkg);
	
	if (opt->append && access(log, F_OK) < 0)
		opt->append = 0;
	
	f = paco_fopen(log, opt->append ? "a" : "w");
	
	if (!opt->append)
		fprintf(f, "#!paco-%s %d\n", PACO_VERSION, (int)time(0));
		
	/* Open '_tmp' and select the valid entries */
	ftmp = paco_fopen(_tmp, "r");	
	while (paco_fgets(buf, sizeof(buf), ftmp)) {
		if (filter_file(buf))
			fprintf(f, "%s\n", zstrip(buf));
	}
	fclose(ftmp);
	
	fclose(f);
	
	if (!opt->append)
		getinfo(opt->pkg);
	else {
		xasprintf(&info_log, "%s/_info/%s", _logdir, opt->pkg);
		if (access(info_log, F_OK) < 0)
			getinfo(opt->pkg);
		free(info_log);
	}
	
	if (!paco_update(log, _tmp, _apparent))
		unique(log);
	
	free(log);
}


void do_log(char *args[], log_opts_t *opt)
{
	/* Get the list of paths to include/exclude */
	if (!opt->include)
		paco_getconf(_pacorc, "INCLUDE", (void *)&opt->include, PACO_STRING, 0);
	if (!opt->include)
		opt->include = strdup(INCLUDE);
	if (!opt->exclude)
		paco_getconf(_pacorc, "EXCLUDE", (void *)&opt->exclude, PACO_STRING, 0);
	if (!opt->exclude)
		opt->exclude = strdup(EXCLUDE);
	
	if (!args)
		log_stdin(opt);
	else
		log_cmd(args, opt);
	
	if (unique(_tmp) == PACO_ERROR_EMPTY_LOG)
		return;
	
	if (opt->pkg)
		log_pkg(opt);
	else
		log_stdout();
}


/* vim: set ts=4: */
