/*********************************************************************
 * paco.h - paco header file
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#ifndef __PACO_H_
#define __PACO_H_

#include "config.h"
#include "util.h"

#ifdef ENABLE_NLS
#	include <libintl.h>
#	define _(x) gettext(x)
#else
#	define _(x) x
#endif

#define ALL_PACKAGES  NULL

#define PACO_VERBOSE  1
#define PACO_DEBUG    2

#define ALLVARS \
	{ "CC", "CFLAGS", "LDFLAGS", "CPPFLAGS", "CXX", "CXXFLAGS", NULL }
	
#define IMAGE_FORMATS  { ".png", ".xpm", ".jpg", ".ico", NULL }


/*=== Typedefs =====================================================*/

typedef struct {
	unsigned long size_inst;
	unsigned long size_miss;
	unsigned long n_inst;
	unsigned long n_miss;
} total_t;


/* List options */
typedef struct {
	
	/* type of files to list */
	int f_type;
#define FILE_INSTALLED	(1 << 0)
#define FILE_MISSING	(1 << 1)
#define FILE_SHARED		(1 << 2)
#define FILE_LOGGED		(FILE_INSTALLED | FILE_MISSING)
	
	/* files list options */
	int f_flags;
#define FILE_SIZE		(1 << 0)
#define FILE_SHARER		(1 << 1)
#define FILE_NO_PKGNAME	(1 << 2)
#define FILE_SYMLINKS	(1 << 3)
	
	/* packages list options */
	int p_flags;
#define PKG_N_INST		(1 << 0)
#define PKG_N_MISS		(1 << 1)
#define PKG_N_SHARED	(1 << 2)
#define PKG_N			(PKG_N_INST | PKG_N_MISS | PKG_N_SHARED)
#define PKG_DATE		(1 << 3)
#define PKG_HOUR		(1 << 4)
#define PKG_SIZE_INST	(1 << 5)
#define PKG_SIZE_MISS	(1 << 6)
#define PKG_SIZE		(PKG_SIZE_INST | PKG_SIZE_MISS)
#define PKG_ONE_COLUMN	(1 << 7)
	
	/* general list options */
	int flags;
#define LIST_REVERSE	(1 << 0)
#define LIST_TOTAL		(1 << 1)
	enum {
		SORT_NAME,
		SORT_SIZE,
		SORT_MSIZE,	
		SORT_N_FILES,
		SORT_N_MFILES,
		SORT_DATE
	} sort;
	long blksize;
} list_opts_t;


/* Remove options */
typedef struct {
	int flags;
#define RM_SHARED	(1 << 0)
#define RM_BATCH	(1 << 1)
#define RM_UNLOG	(1 << 2)
	char *skip;
} rm_opts_t;


/* Log options */
typedef struct {
	int append;
	int ignore_errors;
	char *pkg;
	char *include;
	char *exclude;
} log_opts_t;


/* Singly linked list */
typedef struct list_t {
	struct list_t *next;
	void *data;
} list_t;


/* A file */
typedef struct {
	char *name;
	long size;
	int shared;
} file_t;


/* A package */
typedef struct {
	list_t *files;
	int date;				/* Installation date */
	char *log;				/* Log of the package */
	char *name;				/* Name of the package */
	char *info;				/* Info log of the package */
	unsigned long size_inst;
	unsigned long size_miss;
	unsigned long n_inst;
	unsigned long n_miss;
} pkg_t;


typedef int (list_cmp_f)(list_t *, list_t *);


/*=== Global variables ===============================================*/

extern int _exit_status, _expand, _verbosity, _apparent;
extern char *_pacorc, *_logdir, _tmp[PATH_MAX];


/*=== Global functions ===============================================*/

/* Spit out n times the character c */
#define putnchar(c,n)  do { int x = n; while (x--) putchar(c); } while (0)

/* Avoid atoi(NULL) segfaults */
#define xatoi(s)  ((s) ? atoi(s) : 0)
	
/*-- in util.c --*/

/* Free a char** */
extern void xfree2(char **);

/* Misc. */
extern char **unics(char **in);
extern int get_screen_width();
extern void realloc_strv(char ***strv, char *str);

/* Printing functions */
#define die        paco_die
#define die_errno  paco_die_errno
extern void die_help(const char *fmt, ...)
	__attribute__((__noreturn__,__format__(printf, 1, 2)));
extern void say(const char *fmt, ...)
	__attribute__((__format__(printf, 1, 2)));
extern void say_errno(const char *fmt, ...)
	__attribute__((__format__(printf, 1, 2)));
extern void debug(int print_progname, const char *fmt, ...)
	__attribute__((__format__(printf, 2, 3)));

/* Wrappers around some std functions */
extern void xfree(void *);
extern void *xmalloc0(size_t size);
extern void *xmalloc(size_t size);
extern void *xrealloc(void *p, size_t size);
extern int xasprintf(char **p, const char *fmt, ...)
	__attribute__((__format__(printf, 2, 3)));

extern FILE *tmp_fopen(char *tmpfile, size_t size);

/*-- in paclist.c (singly-linked list stuff) --*/
extern int list_length(list_t *);
extern void get_files(list_t *, int type);
extern void get_totals(list_t *, total_t *total);
extern void list_foreach(list_t *, void (*f)(void *, void *), void *data);
extern list_t *get_pkgs(char **match);
extern list_t *list_reverse(list_t *);
extern list_t *list_sort(list_t *, list_cmp_f);
extern void pkg_list_free(list_t *list);
extern int sort_files_by_size(list_t *p, list_t *q);
extern int sort_files_by_name(list_t *p, list_t *q);
extern void sort_pkg_files_by_name(void *data, __UNUSED void *user_data);

/* Various */
extern void do_log			(char **args, log_opts_t *);
extern void do_list_files	(char **args, list_opts_t *);
extern void do_list_pkgs	(char **args, list_opts_t *);
extern void do_remove		(char **args, rm_opts_t *);
extern void do_var			(char **args, char **var);
extern void do_info			(char **args);
extern void do_conf			(char **args);
extern void do_query		(char **args);
extern void do_unlog		(char **args);
extern void do_update		(char **args);

extern void getinfo(char *pkg_name);

extern int update_log(const char *log, const char *info_log);


#endif  /* __PACO_H_ */


/* vim: set ts=4: */
