/*********************************************************************
 * util.c: Global utilities
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "paco.h"
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>  /* fchmod() */

/* Used to determine the screen width */
#ifdef HAVE_STRUCT_WINSIZE
#	ifdef HAVE_TERMIOS_H
#		include <termios.h>
#	endif
#	ifdef HAVE_SYS_IOCTL_H
#		include <sys/ioctl.h>
#	endif
#	ifndef TIOCGWINSZ
#		define TIOCGWINSZ 0x5413
#	endif
#endif


int get_screen_width()
{
	int width;
#ifdef HAVE_STRUCT_WINSIZE
	struct winsize w;
#endif

	if (!(width = xatoi(getenv("COLUMNS")))) {
		width = 80;
#ifdef HAVE_STRUCT_WINSIZE
		if (isatty(1) && (ioctl(1, TIOCGWINSZ, &w) != -1) && w.ws_col)
			width = w.ws_col;
#endif
	}
	
	return width;
}


void xfree(void *p)
{
	if (p) free(p);
}


/* Free a NULL terminated array of strings. */
void xfree2(char **m)
{
	int i;
	if (m) {
		for (i = 0; m[i]; free(m[i++])) ;
		free(m);
	}
}


/* Remove repeated strings from a NULL terminated array of strings. */
char **unics(char **in)
{
	char **out = NULL;
	int i, j, rep;

	if (!in)
		return NULL;
		
	for (i = 0; in[i]; i++) {
		for (j = 0, rep = 0; (!rep && j < i); j++)
			rep = !strcmp(in[j], in[i]);
		if (rep)
			free(in[i]);
		else
			realloc_strv(&out, in[i]);
	}
	
	return out;
}


void realloc_strv(char ***strv, char *str)
{
	int i = -1;

	if (!*strv) {
		*strv = xmalloc(2 * sizeof(char *));
		i = 0;
	}
	else {
		while (*(*strv + ++i));
		*strv = xrealloc(*strv, (i + 2) * sizeof(char *));
	}
	*(*strv + i++) = strdup(str);
	*(*strv + i) = NULL;
}


void die_help(const char *fmt, ...)
{
	va_list a;
	
	fflush(stdout);
	if (fmt) {
		fprintf(stderr, "paco: ");
		va_start(a, fmt);
		vfprintf(stderr, fmt, a);
		va_end(a);
		fputc('\n', stderr);
	}
	fprintf(stderr, _("Try '%s --help' for more information\n"), "paco");
	
	exit(EXIT_FAILURE);
}


/* Verbose errno message */
void say_errno(const char *fmt, ...)
{
	va_list a;
	
	if (_verbosity == 0)
		return;
	fflush(stdout);
	fprintf(stderr, "paco: ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
	fprintf(stderr, ": %s\n", strerror(errno));
}


/* Verbose message */
void say(const char *fmt, ...)
{
	va_list a;

	if (_verbosity == 0)
		return;
	fflush(stdout);
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
}


/* Debug messages */
void debug(int print_progname, const char *fmt, ...)
{
	va_list a;

	if (_verbosity < PACO_DEBUG)
		return;
	fflush(stdout);
	if (print_progname)
		fprintf(stderr, "paco >> ");
	va_start(a, fmt);
	vfprintf(stderr, fmt, a);
	va_end(a);
}


void *xrealloc(void *ptr, size_t size)
{
	void *p;
	
	if (!ptr)
		return xmalloc(size);
	else if (!(p = realloc(ptr, size)))
		die("realloc(%p, %d): %s", ptr, (int)size, _("Out of memory"));
		
	return p;
}


void *xmalloc(size_t size)
{
	void *ret;
	if (!(ret = malloc(size)))
		die("malloc(\"%d\"): %s", (int)size, _("Out of memory"));
	return ret;
}


void *xmalloc0(size_t size)
{
	void *p = xmalloc(size);
	memset(p, 0, size);
	return p;
}


#if HAVE_VASPRINTF

int xasprintf(char **p, const char *fmt, ...)
{
	va_list a;
	int ret;

	va_start(a, fmt);
	if ((ret = vasprintf(p, fmt, a)) < 0)
		die("vasprintf() %s", _("failed"));
	va_end(a);
		
	return ret;
}

#else

int xasprintf(char **p, const char *fmt, ...)
{
	int n;
	va_list a;

	*p = xmalloc(256);
	
	va_start(a, fmt);
	if ((n = vsnprintf(*p, 256, fmt, a)) < 0)
		die("vsnprintf() %s", _("failed"));
	va_end(a);
	
	if (n >= 256) {
		if (n > (KILOBYTE << 3))
			die("vsnprintf() %s", _("failed"));
		*p = (char *)xrealloc(*p, n);
		va_start(a, fmt);
		if ((n = vsnprintf(*p, n, fmt, a)) < 0)
			die("vsnprintf() %s", _("failed"));
		va_end(a);
	}
	
	return n;
}

#endif  /* HAVE_VASPRINTF */


FILE *tmp_fopen(char *tmpfile, size_t size)
{
	char *tmpdir;
	int fd;

 	tmpdir = getenv("TMPDIR");
	snprintf(tmpfile, size, "%s/pacoXXXXXX", tmpdir ? tmpdir : "/tmp");
	
	if ((fd = mkstemp(tmpfile)) < 0)
		die_errno("mkstemp(\"%s\")", tmpfile);
	
	fchmod(fd, 0644);
	close(fd);
	
	return paco_fopen(tmpfile, "w");
}


/* vim: set ts=4: */
